import dataclasses
import http
import typing

from imbi import automations, errors, models
from imbi.clients import gitlab


@dataclasses.dataclass
class GitLabContext:
    integration_name: str
    project_info: gitlab.ProjectInfo


async def create_project(
    context: automations.AutomationContext,
    automation: models.Automation,
    project: models.Project,
) -> None:
    """Create a GitLab repository for an Imbi project

    This automation creates the GitLab project, adds the ID of the
    new project as a project identifier, and sets the GitLab link
    of we have one configured. The project is created in the GitLab
    group associated with the Project namespace and the sub-path
    associated with the Project Type.

    The gitlab project information is saved in the context using the
    function object (eg, `create_project`) as the key.

    @see https://gitlab.aweber.io/help/api/projects.md#create-project
    """
    client = await gitlab.create_client(context.application,
                                        automation.integration_name,
                                        context.user)
    group_name = (
        (project.namespace.gitlab_group_name or project.namespace.slug),
        (project.project_type.gitlab_project_prefix
         or project.project_type.slug),
    )
    group_info = await client.fetch_group(*group_name)
    if not group_info:
        raise errors.ApplicationError(http.HTTPStatus.BAD_REQUEST,
                                      'misconfigured',
                                      'Group %r does not exist in GitLab',
                                      '/'.join(group_name),
                                      title='Imbi Misconfiguration')

    gitlab_info = await client.create_project(
        group_info,
        project.name,
        description=project.description,
        path=project.slug,
    )
    context[create_project] = GitLabContext(
        integration_name=automation.integration_name, project_info=gitlab_info)
    context.add_callback(delete_project)
    context.note_progress(
        'created GitLab project %s (id=%s) for Imbi project %s',
        gitlab_info.name_with_namespace, gitlab_info.id, project.id)

    await context.run_query(
        'INSERT INTO v1.project_identifiers'
        '            (external_id, integration_name, project_id,'
        '             created_at, created_by)'
        '     VALUES (%(gitlab_id)s, %(integration_name)s, %(imbi_id)s,'
        '             CURRENT_TIMESTAMP, %(username)s)', {
            'gitlab_id': gitlab_info.id,
            'imbi_id': project.id,
            'integration_name': automation.integration_name,
            'username': context.user.username,
        })
    context.note_progress(
        'registered GitLab identifier %s for Imbi project %s', gitlab_info.id,
        project.id)

    settings = context.application.settings['automations']['gitlab']
    link_type_id = settings.get('project_link_type_id', None)
    if link_type_id is not None:
        await context.run_query(
            'INSERT INTO v1.project_links'
            '            (project_id, link_type_id, created_by, url)'
            '     VALUES (%(imbi_id)s, %(link_type_id)s, %(username)s,'
            '             %(project_url)s)', {
                'imbi_id': project.id,
                'link_type_id': link_type_id,
                'project_url': gitlab_info.web_url,
                'username': context.user.username,
            })
        context.note_progress('created GitLab link %r for Imbi project %s',
                              gitlab_info.web_url, project.id)


async def delete_project(context: automations.AutomationContext,
                         _error: BaseException):
    """Compensating action that deletes a GitLab project

    The project information is expected to be in ``context[create_project]``.
    """
    try:
        gitlab_info = typing.cast(GitLabContext, context[create_project])
    except KeyError:
        return

    try:
        client = await gitlab.create_client(context.application,
                                            gitlab_info.integration_name,
                                            context.user)
    except errors.ClientUnavailableError:
        pass
    else:
        await client.delete_project(gitlab_info.project_info)



import http.client
import logging
import typing
import urllib.parse

import pydantic
import sprockets.mixins.http
import tornado.web
import yarl

from imbi import errors, oauth2, user, version
if typing.TYPE_CHECKING:
    from imbi import app


class RepositoryOwner(pydantic.BaseModel):
    login: str


class Repository(pydantic.BaseModel):
    id: int
    name: str
    owner: RepositoryOwner
    html_url: str


class GitHubAPIFailure(errors.ApplicationError):
    def __init__(self, response: sprockets.mixins.http.HTTPResponse,
                 log_message: str, *log_args, **kwargs) -> None:
        status_code = response.code
        kwargs.setdefault('title', 'GitHub API Failure')
        kwargs['github_response_body'] = response.body

        # A GitHub "unauthorized" response is not the same as an
        # Imbi "unauthorized". HTTP unauthorized always refers to
        # the user requested resource which is an Imbi resource.
        if status_code == http.HTTPStatus.UNAUTHORIZED:
            status_code = http.HTTPStatus.FORBIDDEN

        super().__init__(status_code, 'github-api-failure', log_message,
                         *log_args, **kwargs)


async def create_client(
    application: 'app.Application',
    integration_name: str,
    current_user: user.User,
) -> 'GitHubClient':
    tokens = await current_user.fetch_integration_tokens(integration_name)
    if not tokens:
        raise errors.ClientUnavailableError(
            integration_name,
            f'no {integration_name!r} token for {current_user.username}',
        )

    return GitHubClient(current_user, tokens[0], application)


class GitHubClient(sprockets.mixins.http.HTTPClientMixin):
    def __init__(
        self,
        user: user.User,
        token: oauth2.IntegrationToken,
        application: tornado.web.Application,
    ):
        super().__init__()
        self.logger = logging.getLogger(__package__).getChild('GitHubClient')
        self.user = user
        self.token = token
        self.headers = {
            'Accept': str(sprockets.mixins.http.CONTENT_TYPE_JSON),
            'Authorization': f'Bearer {self.token.access_token}',
            'X-GitHub-Api-Version': '2022-11-28',
        }

    async def api(self,
                  url: typing.Union[yarl.URL, str],
                  *,
                  method: str = 'GET',
                  **kwargs):
        """Make an authenticated request to the GitHub API."""
        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        if not url.is_absolute():
            url = self.token.integration.api_endpoint / url.path.lstrip('/')
        request_headers = kwargs.setdefault('request_headers', {})
        request_headers.update(self.headers)
        if kwargs.get('body', None) is not None:
            request_headers['Content-Type'] = str(
                sprockets.mixins.http.CONTENT_TYPE_JSON)
            kwargs['content_type'] = str(
                sprockets.mixins.http.CONTENT_TYPE_JSON)
        kwargs['user_agent'] = f'imbi/{version} (GitHubClient)'
        self.logger.debug('%s %s', method, url)

        response = await super().http_fetch(str(url), method=method, **kwargs)
        if response.code == http.client.BAD_REQUEST and response.body:
            self.logger.warning(
                '%s %s failed: status=%s response=%r',
                method,
                url,
                response.code,
                response.body,
            )
        if response.code == http.client.UNAUTHORIZED:
            await self._refresh_token()
            request_headers.update(self.headers)
            response = await super().http_fetch(str(url),
                                                method=method,
                                                **kwargs)
        return response

    async def _refresh_token(self) -> None:
        params = urllib.parse.urlencode({
            'grant_type': 'refresh_token',
            'refresh_token': self.token.refresh_token,
            'client_id': self.token.integration.client_id,
            'client_secret': self.token.integration.client_secret,
        })
        url = f'{self.token.integration.token_endpoint}?{params}'
        self.logger.info('refreshing token for %s (%s)', self.user.username,
                         self.token.external_id)
        self.logger.debug('refresh_request: %r', params)
        response = await super().http_fetch(
            url,
            method='POST',
            allow_nonstandard_methods=True,
        )
        if response.ok:
            self.logger.debug('refresh response: %r', response.body)
            await self.token.integration.upsert_user_tokens(
                self.user.username,
                self.token.external_id,
                response.body['access_token'],
                response.body['refresh_token'],
                self.token.id_token,
            )
            self.token.access_token = response.body['access_token']
            self.token.refresh_token = response.body['refresh_token']
            self.headers['Authorization'] = f'Bearer {self.token.access_token}'
        else:
            self.logger.error(
                'failed to refresh token for %s: %r',
                self.user.username,
                response.body,
            )
            raise GitHubAPIFailure(response, 'token refresh failed: %s',
                                   response.code)

    async def create_repository(self, org: str, name: str, project_id: int,
                                namespace_slug: str,
                                **attributes) -> Repository:
        """Create a new repository on GitHub.

        Args:
            org: The organization to create the repository in
            name: The name of the repository
            project_id: The Imbi project ID for custom properties
            namespace_slug: The namespace slug for custom properties
            **attributes**: Additional repository attributes

        Docs: https://docs.github.com/en/rest/repos/repos?apiVersion=2022-11-28#create-an-organization-repository
        """  # noqa: E501
        # Set default values for Imbi repositories
        body = {
            'name': name,
            'has_issues': False,
            'has_projects': False,
            'has_wiki': False,
            'auto_init': False,
            'delete_branch_on_merge': True,
            'custom_properties': {
                'imbi_project_id': str(project_id),
                'team': namespace_slug.lower(),
            },
            **attributes
        }

        response = await self.api(
            f'/orgs/{org}/repos',
            method='POST',
            body=body,
        )
        if response.ok:
            return Repository.model_validate(response.body)
        else:
            if response.code == http.HTTPStatus.UNPROCESSABLE_ENTITY:
                self.logger.error(
                    'failed to create repository %s/%s (422): %s',
                    org,
                    name,
                    response.body,
                )
            raise GitHubAPIFailure(
                response,
                'failed to create repository %s',
                name,
            )

    async def delete_repository(self, org: str, repo: str) -> None:
        """Delete a repository on GitHub.

        Docs: https://docs.github.com/en/rest/repos/repos?apiVersion=2022-11-28#delete-a-repository

        Args:
            org: The organization that contains the repository
            repo: The name of the repository to delete
        """  # noqa: E501
        response = await self.api(
            f'/repos/{org}/{repo}',
            method='DELETE',
        )
        if not response.ok or response.code not in (http.HTTPStatus.NOT_FOUND,
                                                    http.HTTPStatus.GONE):
            raise GitHubAPIFailure(
                response,
                '%s failed to delete repository %s',
                org,
                repo,
            )

    async def add_team_to_repository(
        self,
        org: str,
        repo: str,
        team_slug: str,
        permission: str = 'maintain',
    ) -> None:
        """
        Add a team to a repository with specified permissions.

        Args:
            org: The organization that contains the repository
            repo: The repository name
            team_slug: The team slug
            permission: Permission level ('pull', 'triage', 'push', 'maintain')
        """
        payload = {'permission': permission}

        response = await self.api(
            f'/orgs/{org}/teams/{team_slug}/repos/{org}/{repo}',
            method='PUT',
            body=payload,
        )
        if not response.ok:
            raise GitHubAPIFailure(
                response,
                'Failed to add team to repository',
            )

    async def create_environment(
        self,
        org: str,
        repo: str,
        environment: str,
    ) -> None:
        """
        Create an environment for a repository.

        Args:
            org: The organization that contains the repository
            repo: The repository name
            environment: The name of the environment to create
        """
        environment = environment.lower().replace(' ', '-')
        response = await self.api(
            f'/repos/{org}/{repo}/environments/{environment}',
            method='POST',
            body={},
        )
        if not response.ok:
            raise GitHubAPIFailure(
                response,
                'Failed to create environment',
            )

    async def delete_environment(
        self,
        org: str,
        repo: str,
        environment: str,
    ) -> None:
        """
        Delete an environment for a repository.

        Args:
            org: The organization that contains the repository
            repo: The repository name
            environment: The name of the environment to delete
        """
        environment = environment.lower().replace(' ', '-')
        response = await self.api(
            f'/repos/{org}/{repo}/environments/{environment}',
            method='DELETE',
        )
        if not response.ok:
            raise GitHubAPIFailure(
                response,
                'Failed to delete environment',
            )



from __future__ import annotations

import datetime
import http.client
import typing

import jsonpatch
import psycopg2.errors
import pydantic
import sprockets_postgres
from psycopg2 import extensions

from imbi import errors, postgres, slugify
from imbi.endpoints import base
from imbi.automations import models

PathIdType: typing.TypeAlias = typing.Union[int, slugify.Slug]


class AddAutomationRequest(pydantic.BaseModel):
    name: str
    categories: list[models.AutomationCategory] = pydantic.Field(min_length=1)
    callable: models.CallableType
    applies_to: list[PathIdType] = pydantic.Field(default_factory=list,
                                                  min_length=1)
    depends_on: list[PathIdType] = pydantic.Field(default_factory=list)


def adapt_automation_category(category: models.AutomationCategory):
    return extensions.AsIs(repr(category.value))


extensions.register_adapter(models.AutomationCategory,
                            adapt_automation_category)


class CollectionRequestHandler(base.PydanticHandlerMixin,
                               base.ValidatingRequestHandler):
    NAME = 'automations'  # used in metrics

    async def get(self, integration_name: str) -> None:
        result = await self.postgres_execute(
            'SELECT a.id, a.name, a.slug, a.callable, a.categories,'
            '       a.integration_name, a.created_at, a.created_by,'
            '       a.last_modified_at, a.last_modified_by,'
            '       array_agg(DISTINCT pt.slug) AS applies_to,'
            '       array_agg(DISTINCT d.slug) AS depends_on,'
            '       array_agg(DISTINCT pt.id) AS applies_to_ids,'
            '       array_agg(DISTINCT d.id) AS depends_on_ids'
            '  FROM v1.automations AS a'
            '  LEFT JOIN v1.available_automations AS aa'
            '         ON aa.automation_id = a.id'
            '  LEFT JOIN v1.project_types AS pt ON pt.id = aa.project_type_id'
            '  LEFT JOIN v1.automations_graph AS g ON g.automation_id = a.id'
            '  LEFT JOIN v1.automations AS d ON d.id = g.dependency_id'
            ' WHERE a.integration_name = %(integration_name)s'
            ' GROUP BY a.id, a.name, a.slug, a.callable, a.categories,'
            '          a.integration_name, a.slug',
            parameters={'integration_name': integration_name},
            metric_name='get-automations')
        self.send_response(
            [models.Automation.model_validate(row) for row in result])

    @base.require_permission('admin')
    async def post(self, integration_name: str) -> None:
        request = self.parse_request_body_as(AddAutomationRequest)
        slug = slugify.path_element(f'{integration_name}-{request.name}')
        if slug in request.depends_on:
            raise errors.BadRequest('Automation cannot depend on itself')

        self.logger.info('creating automation targeting %r', request.callable)

        async with self.postgres_transaction() as conn:
            conn: sprockets_postgres.PostgresConnector
            result = await conn.execute(
                'INSERT INTO v1.automations(name, slug, integration_name,'
                '                           callable, created_by, categories)'
                '     VALUES (%(name)s, %(slug)s, %(integration_name)s,'
                '             %(callable)s, %(username)s,'
                '             %(categories)s::v1.automation_category_type[])'
                '  RETURNING id, created_at',
                parameters={
                    **request.model_dump(mode='json'),
                    'integration_name': integration_name,
                    'slug': slug,
                    'username': self._current_user.username,
                },
                metric_name='create-automation')
            if not result.row_count:
                raise errors.DatabaseError('No rows returned from INSERT',
                                           title='Failed to insert record')

            automation = models.Automation(
                id=result.row['id'],
                name=request.name,
                slug=slug,
                callable=request.callable,
                categories=request.categories,
                integration_name=integration_name,
                created_at=result.row['created_at'],
                created_by=self._current_user.username,
            )

            if request.applies_to:
                mapping = await slugify.IdSlugMapping.from_database(
                    conn, 'v1', 'project_types', request.applies_to)
                await postgres.insert_values(
                    conn, 'v1', 'available_automations',
                    ['automation_id', 'project_type_id'],
                    [(automation.id, value) for value in mapping.ids])
                automation.applies_to = sorted(mapping.slugs)

            if request.depends_on:
                mapping = await slugify.IdSlugMapping.from_database(
                    conn, 'v1', 'automations', request.depends_on)
                await postgres.insert_values(
                    conn, 'v1', 'automations_graph',
                    ['automation_id', 'dependency_id'],
                    [(automation.id, value) for value in mapping.ids])
                automation.depends_on = sorted(mapping.slugs)

            self.send_response(automation)


class RecordRequestHandler(base.PydanticHandlerMixin,
                           base.ValidatingRequestHandler):
    NAME = 'automation'  # used in metrics

    def _add_self_link(self, _path: str) -> None:
        pass  # this stops base.RequestHandler from making assumptions

    async def get(self, integration_name: str, slug: str) -> None:
        automation = await self._find_automation(integration_name, slug)
        if not automation:
            raise errors.ItemNotFound(instance=self.request.uri)

        path = self.reverse_url('automation', automation.integration_name,
                                automation.slug)
        self.add_header(
            'link',
            f'<{self.request.protocol}://{self.request.host}{path}>;'
            f' rel="self"',
        )
        self.send_response(automation)

    @base.require_permission('admin')
    async def delete(self, integration_name: str, slug: str) -> None:
        automation_id, automation_slug = slugify.decode_path_slug(slug)
        del slug  # use either automation_id or automation_slug

        result = await self.postgres_execute(
            'DELETE FROM v1.automations'
            ' WHERE integration_name = %(integration_name)s'
            '   AND (id = %(automation_id)s'
            '        OR slug = %(automation_slug)s'
            '        OR name = %(automation_slug)s)', {
                'automation_id': automation_id,
                'automation_slug': automation_slug,
                'integration_name': integration_name,
            },
            metric_name='delete-automation')
        if not result.row_count:
            raise errors.ItemNotFound(instance=self.request.uri)
        self.set_status(204, reason='Item Deleted')

    @base.require_permission('admin')
    async def patch(self, integration_name: str, slug: str) -> None:
        patch = jsonpatch.JsonPatch(self.get_request_body())
        automation = await self._find_automation(integration_name, slug)
        if not automation:
            raise errors.ItemNotFound(instance=self.request.uri)

        original = automation.model_dump(mode='json')
        updated = patch.apply(original)
        if updated['id'] != original['id']:
            raise errors.BadRequest('Automation ID is immutable')

        project_type_map, automations_map = await self._get_maps(
            automation, updated)

        # map any IDs into slugs first since that is the real data model
        updated.update({
            'applies_to': project_type_map.to_slugs(updated['applies_to']),
            'depends_on': automations_map.to_slugs(updated['depends_on']),
            'last_modified_at': datetime.datetime.now(datetime.timezone.utc),
            'last_modified_by': self._current_user.username,
        })
        try:
            new_automation = models.Automation.model_validate(updated)
        except pydantic.ValidationError as error:
            raise errors.PydanticValidationError(
                error, 'Failed to validate new Automation') from None

        entity_updated = any(updated[column] != original[column]
                             for column in ('name', 'slug', 'integration_name',
                                            'callable', 'categories'))

        # figure out which association values we are adding and/or
        # removing... and map them to IDs while we are in there
        added_types, removed_types = _partition_map_and_set_changes(
            automation.applies_to, new_automation.applies_to, project_type_map)
        added_deps, removed_deps = _partition_map_and_set_changes(
            automation.depends_on, new_automation.depends_on, automations_map)

        # and now we can finally figure out if anything has changed
        if not any((entity_updated, added_types, removed_types, added_deps,
                    removed_deps)):
            self.set_status(http.client.NOT_MODIFIED)
            return

        async with self.postgres_transaction() as txn:
            txn: sprockets_postgres.PostgresConnector
            if entity_updated:
                await postgres.update_entity(
                    txn, 'v1', 'automations', original, updated,
                    ('name', 'slug', 'integration_name', 'callable',
                     'categories', 'last_modified_at', 'last_modified_by'))
            if added_deps:
                await postgres.insert_values(
                    txn, 'v1', 'automations_graph',
                    ['automation_id', 'dependency_id'],
                    [(automation.id, dependency) for dependency in added_deps])
            if removed_deps:
                await txn.execute(
                    'DELETE FROM v1.automations_graph'
                    '      WHERE automation_id = %(automation_id)s'
                    '        AND dependency_id IN %(removed_deps)s', {
                        'automation_id': automation.id,
                        'removed_deps': tuple(removed_deps)
                    },
                    metric_name='delete-automations_graph')
            if added_types:
                await postgres.insert_values(
                    txn, 'v1', 'available_automations',
                    ['automation_id', 'project_type_id'],
                    [(automation.id, pt_id) for pt_id in added_types])
            if removed_types:
                await txn.execute(
                    'DELETE FROM v1.available_automations'
                    '      WHERE automation_id = %(automation_id)s'
                    '        AND project_type_id IN %(removed_types)s', {
                        'automation_id': automation.id,
                        'removed_types': tuple(removed_types)
                    },
                    metric_name='delete-available_automations')

        automation = await self._find_automation(integration_name, slug)
        self.send_response(automation)

    def on_postgres_error(self, metric_name: str,
                          exc: Exception) -> typing.Optional[Exception]:
        """Handle postgres exceptions

        Stop exceptions that are not handled in sprockets_postgres
        from becoming "internal server errors". we don't want a bad
        patch to be retried by a client.

        """
        self.logger.exception('unexpected error')
        if isinstance(exc, psycopg2.errors.NotNullViolation):
            sanitized = exc.pgerror.removeprefix('ERROR:').splitlines()[0]
            return errors.BadRequest(
                'Request generated an invalid result: %s',
                str(exc).replace('\n', ' '),
                title='Invalid update',
                reason='Bad Request',
                detail=sanitized.strip(),
                sqlstate=exc.pgcode,
            )
        if isinstance(exc, psycopg2.errors.CheckViolation):
            sanitized = exc.pgerror.removeprefix('ERROR:').splitlines()[0]
            return errors.ApplicationError(
                http.HTTPStatus.CONFLICT,
                'conflict',
                'Request violated relationship constraint: %s',
                str(exc).replace('\n', ' '),
                title='Invalid update',
                reason='Conflict',
                detail=sanitized.strip(),
                sqlstate=exc.pgcode,
            )
        return super().on_postgres_error(metric_name, exc)

    async def _find_automation(self, integration_name: str,
                               slug: str) -> models.Automation | None:
        """Retrieve a single automation from the DB

        The `slug` can be a slug, the name of the automation,
        or the automation's surrogate ID.

        """
        automation_id, automation_slug = slugify.decode_path_slug(slug)
        del slug  # use either automation_id or automation_slug

        self.logger.debug('looking for %r or %r', automation_id,
                          automation_slug)
        result = await self.postgres_execute(
            'SELECT a.id, a.name, a.callable, a.categories, a.slug,'
            '       a.integration_name, a.created_at, a.created_by,'
            '       a.last_modified_at, a.last_modified_by,'
            '       array_agg(DISTINCT pt.slug) AS applies_to,'
            '       array_agg(DISTINCT d.slug) AS depends_on,'
            '       array_agg(DISTINCT pt.id) AS applies_to_ids,'
            '       array_agg(DISTINCT d.id) AS depends_on_ids'
            '  FROM v1.automations AS a'
            '  LEFT JOIN v1.available_automations AS aa'
            '         ON aa.automation_id = a.id'
            '  LEFT JOIN v1.project_types AS pt ON pt.id = aa.project_type_id'
            '  LEFT JOIN v1.automations_graph AS g ON g.automation_id = a.id'
            '  LEFT JOIN v1.automations AS d ON d.id = g.dependency_id'
            ' WHERE a.integration_name = %(integration_name)s'
            '   AND (a.slug = %(slug)s'
            '        OR a.id = %(automation_id)s'
            '        OR a.name = %(slug)s)'
            ' GROUP BY a.id, a.name, a.callable, a.categories, a.slug,'
            '          a.integration_name', {
                'automation_id': automation_id,
                'slug': automation_slug,
                'integration_name': integration_name,
            },
            metric_name='get-automation')
        if result.row:
            try:
                automation = models.Automation.model_validate(result.row)
            except pydantic.ValidationError as error:
                raise errors.PydanticValidationError(
                    error,
                    'Database contains an invalid automation with id %r: %s',
                    result.row['id'],
                    error,
                    status_code=500)
            self.logger.debug('found %s (%s)', automation.slug, automation.id)
            return automation
        return None

    async def _get_maps(
            self, automation: models.Automation, updated: dict
    ) -> tuple[slugify.IdSlugMapping, slugify.IdSlugMapping]:
        async with self.application.postgres_connector(
                self.on_postgres_error, self.on_postgres_timing) as conn:
            values = set(automation.applies_to)
            values.update(updated.get('applies_to', []))
            project_type_map = await slugify.IdSlugMapping.from_database(
                conn, 'v1', 'project_types', values)

            values = set(automation.depends_on)
            values.update(updated.get('depends_on', []))
            automations_map = await slugify.IdSlugMapping.from_database(
                conn, 'v1', 'automations', values)

        return project_type_map, automations_map


def _partition_map_and_set_changes(
        start_data: list[str], final_data: list[str],
        mapping: slugify.IdSlugMapping) -> tuple[list[int], list[int]]:
    """Take original & changed collections and return added & removed sets"""
    start_set, final_set = set(start_data), set(final_data)
    return (mapping.to_ids(final_set - start_set),
            mapping.to_ids(start_set - final_set))



import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):

    NAME = 'project-fact-types'
    ID = 'project_id'
    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT c.id, c.name, c.fact_type, c.data_type, c.description,
               c.ui_options, c.weight,
               CASE WHEN c.fact_type = 'enum' THEN ARRAY(
                         SELECT value
                           FROM v1.project_fact_type_enums
                          WHERE fact_type_id = c.id
                       ORDER BY value ASC)
                    ELSE NULL END AS enum_values,
               CASE WHEN c.fact_type = 'range' THEN (
                         SELECT min(min_value)
                           FROM v1.project_fact_type_ranges
                          WHERE fact_type_id = c.id)
                  ELSE NULL END AS min_value,
               CASE WHEN c.fact_type = 'range' THEN (
                         SELECT max(max_value)
                           FROM v1.project_fact_type_ranges
                          WHERE fact_type_id = c.id)
                  ELSE NULL END AS max_value
          FROM v1.projects AS a
          JOIN v1.project_types AS b
            ON b.id = a.project_type_id
          JOIN v1.project_fact_types AS c
            ON b.id = ANY (c.project_type_ids)
         WHERE a.id = %(project_id)s
      ORDER BY c.name""")



from __future__ import annotations

import collections.abc
import re
import typing

import pydantic
import sprockets_postgres
from psycopg2 import sql

from imbi import errors

Slug = typing.Annotated[str,
                        pydantic.StringConstraints(
                            pattern=r'^[a-z0-9][-_a-z0-9]*[a-z0-9]?$',
                            min_length=1,
                        )]


class InvalidSlugError(errors.BadRequest):
    def __init__(self, slug_type: str, invalid_slugs: set[int | str]) -> None:
        super().__init__('Invalid %s slug(s): %r',
                         slug_type,
                         invalid_slugs,
                         invalid_slugs=sorted(invalid_slugs))


def path_element(value: str) -> str:
    """Generate a slug for an arbitrary string"""
    value = re.sub(r'[\s_]+', '-', value)
    value = re.sub(r'[^-a-z0-9]', '', value.lower())
    return Slug(value.strip('-_'))


def decode_path_slug(slug_or_id: str) -> tuple[int | None] | tuple[None | str]:
    """Decode a path parameter as either an integer or a slug

    :returns: a tuple of (int, str) where only one of the values
        is not `None`

    """
    id_value: int | None = None
    slug_value: str | None = slug_or_id
    try:
        id_value = int(slug_or_id, 10)
        slug_value = None
    except ValueError:
        pass
    return id_value, slug_value


class IdSlugMapping:
    """Maps between IDs and slugs"""
    def __init__(self, slug_to_id: dict[str, int] | None = None) -> None:
        self._slug_to_id: dict[str, int] = {}
        if slug_to_id:
            self._slug_to_id.update(slug_to_id)
        self._id_to_slug = {v: k for k, v in self._slug_to_id.items()}

    def to_ids(self, values: collections.abc.Iterable[str | int]) -> list[int]:
        """Map `values` to IDs

        This maps non-ID values to their associated ID while
        maintaining any IDs in `values`.

        """
        return [self._slug_to_id.get(v, v) for v in values]

    def to_slugs(self,
                 values: collections.abc.Iterable[str | int]) -> list[str]:
        """Map `values` to slugs

        This maps non-slug values to their associated slug
        while maintaining any slugs in `values`.

        """
        return [self._id_to_slug.get(v, v) for v in values]

    @property
    def ids(self) -> list[int]:
        """List of IDs"""
        return list(self._id_to_slug)

    @property
    def slugs(self) -> list[str]:
        """List of slugs"""
        return list(self._slug_to_id)

    @classmethod
    async def from_database(
            cls, conn: sprockets_postgres.PostgresConnector, schema: str,
            entity_table: str,
            filter_values: collections.abc.Collection[int | str]
    ) -> typing.Self:
        """Build a mapping based on the slug & id values in a table."""
        ids: tuple[int, ...] = ()
        slugs: tuple[str, ...] = ()
        params = {}
        query = sql.SQL('SELECT DISTINCT slug, id FROM {}').format(
            sql.Identifier(schema, entity_table))
        if filter_values:
            empty = (None, )
            ids = tuple(i for i in filter_values if isinstance(i, int))
            slugs = tuple(s for s in filter_values if isinstance(s, str))
            params.update({'ids': ids or empty, 'slugs': slugs or empty})
            query += sql.SQL('WHERE slug IN {} OR id IN {}').format(
                sql.Placeholder('slugs'), sql.Placeholder('ids'))
        result = await conn.execute(query,
                                    params,
                                    metric_name=f'fetch-{entity_table}-slugs')

        mapping = {row['slug']: row['id'] for row in result}
        invalid_values: set[int | str] = set(slugs) - set(mapping.keys())
        invalid_values.update(set(ids) - set(mapping.values()))
        if invalid_values:
            raise InvalidSlugError(entity_table, invalid_values)

        return cls(mapping)



import re

from imbi.endpoints import base


class RequestHandler(base.AuthenticatedRequestHandler):

    NAME = 'reports-namespace-shs-history'

    SQL = re.sub(
        r'\s+', ' ', """\
        SELECT a.namespace_id,
               a.scored_on,
               a.health_score
          FROM v1.namespace_kpi_history AS a
          JOIN v1.namespaces AS b
            ON b.id = a.namespace_id
         WHERE a.scored_on > CURRENT_DATE - interval '1 year'
      ORDER BY a.scored_on ASC, a.namespace_id ASC;""")

    async def get(self):
        result = await self.postgres_execute(self.SQL, metric_name=self.NAME)
        self.send_response(result.rows)



from __future__ import annotations

import collections
import typing

from imbi.endpoints.project_sbom import models


class DependencyGraph:
    _table: dict[models.BOMRef, set[models.BOMRef]]

    def __init__(self,
                 dependencies: typing.Iterable[models.Dependency]) -> None:
        self._table = collections.defaultdict(set)
        for dependency in dependencies:
            self._table[dependency.ref].update(dependency.depends_on)

    def all_dependencies(
            self, from_node: models.BOMRef) -> typing.Iterable[models.BOMRef]:
        visited: set[models.BOMRef] = set()
        queue = list(self._table[from_node])
        while queue:
            elm = queue.pop()
            if elm not in visited:
                yield elm
                visited.add(elm)
                queue.extend(r for r in self._table[elm] if r not in visited)



import re

from imbi.endpoints import base, projects


class _LinkRequestMixin:

    ID_KEY = ['project_id', 'link_type_id']
    ITEM_NAME = 'project-link'
    FIELDS = ['project_id', 'link_type_id', 'url']
    TTL = 300

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT a.project_id,
               a.link_type_id,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               b.link_type,
               b.icon_class,
               a.url
          FROM v1.project_links AS a
          JOIN v1.project_link_types AS b
            ON b.id = a.link_type_id
         WHERE a.project_id = %(project_id)s
           AND a.link_type_id = %(link_type_id)s""")


class CollectionRequestHandler(projects.ProjectAttributeCollectionMixin,
                               _LinkRequestMixin,
                               base.CollectionRequestHandler):

    NAME = 'project-links'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT a.project_id,
               a.link_type_id,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               b.link_type,
               b.icon_class,
               a.url
          FROM v1.project_links AS a
          JOIN v1.project_link_types AS b
            ON b.id = a.link_type_id
         WHERE a.project_id = %(project_id)s
         ORDER BY b.link_type""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.project_links
                    (project_id, link_type_id, created_by, url)
             VALUES (%(project_id)s, %(link_type_id)s, %(username)s, %(url)s)
          RETURNING project_id, link_type_id""")


class RecordRequestHandler(projects.ProjectAttributeCRUDMixin,
                           _LinkRequestMixin, base.CRUDRequestHandler):

    NAME = 'project-link'

    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.project_links
              WHERE project_id = %(project_id)s
                AND link_type_id = %(link_type_id)s""")

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.project_links
           SET url=%(url)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s
         WHERE project_id = %(project_id)s
           AND link_type_id = %(current_link_type_id)s""")



import unittest
import uuid

from tornado import testing

from imbi import timestamp, user
from tests import base


class GroupTestCase(unittest.TestCase):
    def test_as_dict(self):
        name = str(uuid.uuid4())
        permissions = [str(uuid.uuid4()), str(uuid.uuid4())]
        self.assertDictEqual(dict(user.Group(name, permissions)), {
            'name': name,
            'permissions': sorted(permissions)
        })

    def test_repr(self):
        name = str(uuid.uuid4())
        permissions = [str(uuid.uuid4()), str(uuid.uuid4())]
        group = user.Group(name, permissions)
        self.assertEqual(
            repr(group),
            '<Group name={} permissions={}>'.format(name, sorted(permissions)))


class InternalTestCase(base.TestCase):

    SQL_INSERT_GROUP = """\
    INSERT INTO v1.groups ("name", permissions)
         VALUES (%(name)s, %(permissions)s)"""

    SQL_INSERT_GROUP_MEMBERSHIP = """\
    INSERT INTO v1.group_members ("group", username)
         VALUES (%(group)s, %(username)s)"""

    SQL_INSERT_USER = """\
    INSERT INTO v1.users (username, display_name, email_address, password,
                          last_seen_at)
         VALUES (%(username)s, %(display_name)s, %(email_address)s,
                 %(password)s, %(last_seen_at)s)"""

    SQL_UPDATE_DISPLAY_NAME = """\
    UPDATE v1.users SET display_name = %(display_name)s
     WHERE username = %(username)s"""

    async def setup_group(self, values=None):
        if not values:
            values = {}
        values.setdefault('name', str(uuid.uuid4()))
        values.setdefault('permissions',
                          sorted([str(uuid.uuid4()),
                                  str(uuid.uuid4())]))
        await self.postgres_execute(self.SQL_INSERT_GROUP, values)
        return values

    async def setup_group_membership(self, username, group):
        await self.postgres_execute(self.SQL_INSERT_GROUP_MEMBERSHIP, {
            'group': group,
            'username': username
        })

    async def setup_user(self, values=None):
        if not values:
            values = {}
        values.setdefault('username', str(uuid.uuid4()))
        values.setdefault('display_name', str(uuid.uuid4()))
        values.setdefault('email_address',
                          '{}@{}'.format(str(uuid.uuid4()), str(uuid.uuid4())))
        values.setdefault('last_seen_at', timestamp.utcnow())
        password = values.get('password', str(uuid.uuid4()))
        values['password'] = self._app.hash_password(password)
        await self.postgres_execute(self.SQL_INSERT_USER, values)
        values['password'] = password
        return values

    @testing.gen_test
    async def test_authenticate_happy_path(self):
        user_value = await self.setup_user()
        group_value = await self.setup_group()
        await self.setup_group_membership(user_value['username'],
                                          group_value['name'])

        obj = user.User(self._app, user_value['username'],
                        user_value['password'])
        self.assertTrue(await obj.authenticate())

        values = obj.as_dict()
        for key in {
                'created_at', 'last_refreshed_at', 'last_seen_at', 'password'
        }:
            self.assertIn(key, values)
            del values[key]

        expectation = {
            'username': user_value['username'],
            'user_type': 'internal',
            'external_id': None,
            'google_user': False,
            'display_name': user_value['display_name'],
            'email_address': user_value['email_address'],
            'groups': [group_value['name']],
            'permissions': sorted(set(group_value['permissions'])),
            'integrations': [],
        }
        self.assertDictEqual(values, expectation)

        for role in group_value['permissions']:
            self.assertTrue(obj.has_permission(role))
        self.assertFalse(obj.has_permission('other'))

    @testing.gen_test
    async def test_refresh(self):
        user_value = await self.setup_user()
        group_value = await self.setup_group()
        await self.setup_group_membership(user_value['username'],
                                          group_value['name'])

        obj = user.User(self._app, user_value['username'],
                        user_value['password'])
        self.assertTrue(await obj.authenticate())

        display_name = str(uuid.uuid4())
        self.assertNotEqual(display_name, user_value['display_name'])

        await self.postgres_execute(self.SQL_UPDATE_DISPLAY_NAME, {
            'display_name': display_name,
            'username': user_value['username']
        })

        await obj.refresh()

        values = obj.as_dict()
        for key in {
                'created_at', 'last_refreshed_at', 'last_seen_at', 'password'
        }:
            self.assertIn(key, values)
            del values[key]

        expectation = {
            'username': user_value['username'],
            'user_type': 'internal',
            'external_id': None,
            'google_user': False,
            'display_name': display_name,
            'email_address': user_value['email_address'],
            'groups': [group_value['name']],
            'permissions': sorted(set(group_value['permissions'])),
            'integrations': [],
        }
        self.assertDictEqual(values, expectation)

    @testing.gen_test
    async def test_reset_on_authentication_failure(self):
        user_value = await self.setup_user()
        obj = user.User(self._app, user_value['username'],
                        user_value['password'])
        self.assertTrue(await obj.authenticate())
        for key in {'display_name', 'email_address'}:
            self.assertEqual(user_value[key], getattr(obj, key))
        obj._password = str(uuid.uuid4())
        self.assertFalse(await obj.authenticate())
        for key in {'display_name', 'email_address'}:
            self.assertIsNone(getattr(obj, key))

    @testing.gen_test
    async def test_token_authentication(self):
        user_value = await self.setup_user()
        token = await self.get_token(user_value['username'])
        obj = user.User(self._app, None, None, token)
        self.assertTrue(await obj.authenticate())
        for key in {'display_name', 'email_address'}:
            self.assertEqual(user_value[key], getattr(obj, key))

    @testing.gen_test
    async def test_token_authentication_failure(self):
        obj = user.User(self._app, 'foo', None, str(uuid.uuid4()))
        self.assertFalse(await obj.authenticate())

    @testing.gen_test
    async def test_garbage_token_fails(self):
        obj = user.User(self._app, 'foo', None, 'ohno')
        self.assertFalse(await obj.authenticate())

    @testing.gen_test
    async def test_should_refresh_false(self):
        user_value = await self.setup_user()
        obj = user.User(self._app, user_value['username'],
                        user_value['password'])
        self.assertTrue(await obj.authenticate())
        self.assertFalse(obj.should_refresh)

    @testing.gen_test
    async def test_should_refresh_true(self):
        user_value = await self.setup_user()
        obj = user.User(self._app, user_value['username'],
                        user_value['password'])
        self.assertTrue(await obj.authenticate())
        self.assertFalse(obj.should_refresh)
        obj.last_refreshed_at = timestamp.utcnow() - (obj.REFRESH_AFTER * 2)
        self.assertTrue(obj.should_refresh)

    @testing.gen_test
    async def test_update_last_seen_at(self):
        user_value = await self.setup_user()
        obj = user.User(self._app, user_value['username'],
                        user_value['password'])
        self.assertTrue(await obj.authenticate())
        last_seen_at = obj.last_seen_at
        await obj.update_last_seen_at()
        self.assertGreater(obj.last_seen_at, last_seen_at)


class LDAPTestCase(base.TestCase):

    LDAP_USERNAME = 'test'
    LDAP_PASSWORD = 'password'

    @staticmethod
    def get_user_dn(conn):
        result = conn.extend.standard.who_am_i()
        return result[3:]

    @testing.gen_test
    async def test_authenticate(self):
        obj = user.User(self._app, self.LDAP_USERNAME, self.LDAP_PASSWORD)
        self.assertTrue(await obj.authenticate())

        await obj.refresh()

        values = obj.as_dict()
        for key in {
                'created_at', 'last_refreshed_at', 'last_seen_at', 'password'
        }:
            self.assertIn(key, values)
            del values[key]

        expectation = {
            'username': 'test',
            'user_type': 'ldap',
            'external_id': 'cn=test,ou=users,dc=example,dc=org',
            'google_user': False,
            'display_name': 'Its Imbi',
            'email_address': 'imbi@example.org',
            'groups': ['admin', 'imbi'],
            'permissions': ['admin', 'reader'],
            'integrations': [],
        }
        self.assertDictEqual(values, expectation)
        self.assertTrue(obj.has_permission('admin'))
        self.assertFalse(obj.has_permission('other'))

    @testing.gen_test
    async def test_authentication_failure_without_password(self):
        obj = user.User(self._app, self.LDAP_USERNAME)
        self.assertFalse(await obj.authenticate())

    @testing.gen_test
    async def test_reset_on_authentication_failure(self):
        obj = user.User(self._app, self.LDAP_USERNAME, self.LDAP_PASSWORD)
        self.assertTrue(await obj.authenticate())
        for key in {'display_name', 'email_address', 'external_id'}:
            self.assertIsNotNone(getattr(obj, key))
        obj = user.User(self._app, self.LDAP_USERNAME, str(uuid.uuid4()))
        self.assertFalse(await obj.authenticate())
        for key in {'display_name', 'email_address', 'external_id'}:
            self.assertIsNone(getattr(obj, key))

    @testing.gen_test
    async def test_token_authentication(self):
        obj1 = user.User(self._app, self.LDAP_USERNAME, self.LDAP_PASSWORD)
        self.assertTrue(await obj1.authenticate())
        token = await self.get_token(self.LDAP_USERNAME)
        obj2 = user.User(self._app, None, None, token)
        self.assertTrue(await obj2.authenticate())
        self.assertEqual(obj1.username, obj2.username)

    @testing.gen_test
    async def test_should_refresh_false(self):
        obj = user.User(self._app, self.LDAP_USERNAME, self.LDAP_PASSWORD)
        self.assertTrue(await obj.authenticate())
        self.assertFalse(obj.should_refresh)

    @testing.gen_test
    async def test_should_refresh_true(self):
        obj = user.User(self._app, self.LDAP_USERNAME, self.LDAP_PASSWORD)
        self.assertTrue(await obj.authenticate())
        self.assertFalse(obj.should_refresh)
        obj.last_refreshed_at = timestamp.utcnow() - (obj.REFRESH_AFTER * 2)
        self.assertTrue(obj.should_refresh)



import asyncio
import datetime
import re

from tornado import web

from imbi.endpoints import base


class CollectionRequestHandler(base.TimeBasedPaginationMixin,
                               base.ValidatingRequestHandler):

    NAME = 'project-activity-feed'

    PROJECT_FEED_SQL = re.sub(
        r'\s+', ' ', """
        WITH created AS (
          SELECT created_at AS "when", created_by AS who,
                 'created' AS what, NULL AS fact_name, NULL AS value
            FROM v1.projects
           WHERE id = %(project_id)s
        ),
        updated AS (
          SELECT last_modified_at AS "when", last_modified_by AS who,
                 'updated' AS what, NULL AS fact_name, NULL AS value
            FROM v1.projects
           WHERE id = %(project_id)s
        ),
        facts AS (
          SELECT a.recorded_at AS "when", a.recorded_by AS who,
                 'updated fact' AS what, b.name AS fact_name, a.value
            FROM v1.project_facts AS a
            JOIN v1.project_fact_types AS b
              ON b.id = a.fact_type_id
           WHERE a.project_id = %(project_id)s
        ),
        identifiers AS (
          SELECT COALESCE(a.last_modified_at, a.created_at) AS "when",
                 COALESCE(a.last_modified_by, a.created_by) AS who,
                 'updated fact' AS what,
                 (a.integration_name || ' external ID') AS fact_name,
                 a.external_id AS value
            FROM v1.project_identifiers AS a
           WHERE a.project_id = %(project_id)s
        ),
        combined AS (
                SELECT * FROM created
          UNION SELECT * FROM updated
          UNION SELECT * FROM facts
          UNION SELECT * FROM identifiers
        )
        SELECT 'ProjectFeedEntry' AS "type", c."when", c.who, u.display_name,
               c.what, c.fact_name, c.value
          FROM combined AS c
          JOIN v1.users AS u ON u.username = c.who
         WHERE c."when" >  %(earlier)s
           AND c."when" <= %(later)s
         ORDER BY c."when" DESC
         LIMIT %(remaining)s
        """)

    OPERATIONS_LOG_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT o.id, o.recorded_at, o.recorded_by, o.completed_at,
               o.project_id, o.environment, o.change_type, o.description,
               o.link, o.notes, o.ticket_slug, o.version,
               COALESCE(
                  u.email_address,
                  COALESCE(u2.email_address, 'UNKNOWN')
               ) AS email_address,
               COALESCE(
                  u.display_name,
                  COALESCE(u2.display_name, o.recorded_by)
               ) AS display_name,
               p.name AS project_name,
               'OperationsLogEntry' AS "type",
               o.occurred_at,
               o.performed_by
          FROM v1.operations_log AS o
          JOIN v1.projects AS p ON p.id = o.project_id
          LEFT JOIN v1.users AS u ON u.username = o.performed_by
          LEFT JOIN v1.users AS u2 ON u2.username = o.recorded_by
         WHERE p.id = %(project_id)s
           AND recorded_at >  %(earlier)s
           AND recorded_at <= %(later)s
         ORDER BY o.recorded_at DESC, o.id DESC
         LIMIT %(remaining)s
        """)

    EARLIEST_EVENT_SQL = re.sub(
        r'\s+', ' ', """\
        WITH T AS (
          SELECT created_at AS earliest
            FROM v1.projects
           WHERE id = %(project_id)s
          UNION
          SELECT MIN(recorded_at) AS earliest
            FROM v1.operations_log
           WHERE project_id = %(project_id)s
        )
        SELECT COALESCE(MIN(earliest), CURRENT_TIMESTAMP) AS earliest
          FROM T
        """)

    @web.authenticated
    async def get(self, project_id: str) -> None:
        token = await self._get_pagination_token(project_id)
        buckets = await self.fetch_items(token,
                                         self._retrieve_rows,
                                         datetime.timedelta(days=90),
                                         project_id=project_id)
        self.send_response(buckets)

    async def _retrieve_rows(
            self, params: dict) -> list[tuple[datetime.datetime, dict]]:
        activity_feed, ops_log = await asyncio.gather(
            self.postgres_execute(self.PROJECT_FEED_SQL, params),
            self.postgres_execute(self.OPERATIONS_LOG_SQL, params))
        self.logger.debug('fetched page containing %s activity, %s ops log',
                          len(activity_feed), len(ops_log))

        all_rows = [(row['when'], row) for row in activity_feed]
        all_rows.extend((row['recorded_at'], row) for row in ops_log)

        return all_rows

    async def _get_pagination_token(
            self, project_id) -> base.TimeBasedPaginationToken:
        token = self.get_pagination_token_from_request()
        if token is None:
            result = await self.postgres_execute(self.EARLIEST_EVENT_SQL,
                                                 {'project_id': project_id})
            token = base.TimeBasedPaginationToken(
                start=datetime.datetime.now(datetime.timezone.utc),
                limit=int(self.get_query_argument('limit', '25')),
                earliest=result.row['earliest'],
            )
        return token



from __future__ import annotations

import logging

import sprockets_postgres
import typing_extensions as typing

import imbi.models
import imbi.opensearch.project
import imbi.semver
from imbi.endpoints.components import models

if typing.TYPE_CHECKING:
    import imbi.app


async def update_component_version_statuses(
    component: models.Component,
    connector: sprockets_postgres.PostgresConnector,
) -> None:
    if (component.status == models.ComponentStatus.DEPRECATED
            or component.status == models.ComponentStatus.FORBIDDEN
            or component.active_version is None):
        new_status = (component.status.value if component.status
                      in (models.ComponentStatus.DEPRECATED,
                          models.ComponentStatus.FORBIDDEN) else
                      models.ProjectComponentStatus.UNSCORED.value)
        await connector.execute(
            'UPDATE v1.component_versions'
            '   SET status = %(new_status)s'
            ' WHERE package_url = %(package_url)s', {
                'package_url': component.package_url,
                'new_status': new_status
            },
            metric_name=f'bulk-update-component-version-{new_status}')
        return

    transaction: sprockets_postgres.PostgresConnector
    async with connector.transaction() as transaction:
        result = await transaction.execute(
            'SELECT id, version, status'
            '  FROM v1.component_versions'
            ' WHERE package_url = %(package_url)s'
            '   FOR UPDATE', {'package_url': component.package_url})
        for row in result:
            new_status = (models.ProjectComponentStatus.UP_TO_DATE.value
                          if row['version'] in component.active_version else
                          models.ProjectComponentStatus.OUTDATED.value)
            await transaction.execute(
                'UPDATE v1.component_versions'
                '   SET status = %(new_status)s'
                '  WHERE id = %(version_id)s'
                ' AND status = %(old_status)s', {
                    'new_status': new_status,
                    'old_status': row['status'],
                    'version_id': row['id']
                },
                metric_name='update-component-version-status')


async def update_component_score_for_project(
        project_id: int,
        connector: sprockets_postgres.PostgresConnector,
        app: imbi.app.Application,
        *,
        index_project: bool = True):
    logger = logging.getLogger(__package__).getChild(
        'update_component_score_for_project')

    fact_id = app.settings['component_scoring']['project_fact_type_id']
    if not fact_id:
        logger.debug('skipping component score update, scoring fact not set')
        return

    transaction: sprockets_postgres.PostgresConnector
    async with connector.transaction() as transaction:
        result = await transaction.execute(
            'SELECT value::INTEGER'
            '  FROM v1.project_facts'
            ' WHERE project_id = %(project_id)s'
            '   AND fact_type_id = %(fact_id)s', {
                'project_id': project_id,
                'fact_id': fact_id
            })
        current_status = result.row['value'] if result else None

        result = await transaction.execute(
            'SELECT v.status, COUNT(v.id) AS num_components'
            '  FROM v1.project_components AS p'
            '  JOIN v1.component_versions AS v'
            '    ON v.package_url = p.package_url AND p.version_id = v.id'
            ' WHERE p.project_id = %(project_id)s'
            '   AND v.status != %(unscored)s'
            ' GROUP BY v.status', {
                'project_id': project_id,
                'unscored': models.ProjectComponentStatus.UNSCORED.value,
            },
            metric_name='retrieve-tracked-components')

        counts = {name: 0 for name in models.ProjectComponentStatus}
        counts.update({row['status']: row['num_components'] for row in result})
        status = models.ProjectStatus.calculate(counts)

        if status.value != current_status:
            logger.info(
                'updating component score from %r to %r for project %s',
                current_status or 'NULL', status.value, project_id)
            await transaction.execute(
                'INSERT INTO v1.project_facts(project_id, fact_type_id, value,'
                '                             recorded_at, recorded_by)'
                '     VALUES (%(project_id)s, %(fact_id)s, %(value)s,'
                '             CURRENT_TIMESTAMP, %(recorded_by)s)'
                ' ON CONFLICT (project_id, fact_type_id)'
                '   DO UPDATE'
                '         SET value = %(value)s,'
                '             recorded_at = CURRENT_TIMESTAMP,'
                '             recorded_by = %(recorded_by)s', {
                    'project_id': project_id,
                    'fact_id': fact_id,
                    'value': status,
                    'recorded_by': 'system'
                },
                metric_name='set-component-score')

            if index_project:
                project = await imbi.models.project(project_id, app)
                if project is not None:  # may have been removed
                    index = imbi.opensearch.project.ProjectIndex(app)
                    await index.index_document(project)


async def enable_component_scoring(app: imbi.app.Application) -> None:
    """Enable the project component scoring feature

    1. creates the database rows for the fact type
    2. updates app.settings[component_scoring]

    """
    logger = logging.getLogger(__package__).getChild(
        'enable_component_scoring')

    app_config = app.settings.setdefault('component_scoring', {})
    if app_config['enabled']:  # nothing to do here
        return

    connector: sprockets_postgres.PostgresConnector
    transaction: sprockets_postgres.PostgresConnector
    async with app.postgres_connector() as connector:
        async with connector.transaction() as transaction:
            # Create a new project fact using the configured name
            fact_name = app_config['fact_name']
            logger.info('Creating %r project fact type', fact_name)
            result = await transaction.execute(
                'INSERT INTO v1.project_fact_types('
                '               created_by, name, fact_type, data_type,'
                '               project_type_ids, ui_options)'
                '     VALUES (%(created_by)s, %(name)s, %(fact_type)s,'
                '             %(data_type)s, %(empty)s, %(ui_options)s)'
                '  RETURNING id', {
                    'created_by': 'system',
                    'data_type': 'integer',
                    'empty': [],
                    'fact_type': 'range',
                    'name': fact_name,
                    'ui_options': ['read-only'],
                })

            # Create range values that encompass the ProjectStatus enum
            # values in imbi.endpoints.components.models. Note that we start
            # at zero and use a value slightly greater than that upper bound
            # of the last range as the start of the next range.
            app_config['project_fact_type_id'] = result.row['id']
            lower_bound = 0.0
            for upper_bound in sorted(models.ProjectStatus):
                await transaction.execute(
                    'INSERT INTO v1.project_fact_type_ranges('
                    '               created_by, fact_type_id, min_value,'
                    '               max_value, score)'
                    '     VALUES (%(created_by)s, %(fact_type_id)s,'
                    '             %(min_value)s, %(max_value)s,'
                    '             %(max_value)s)', {
                        'created_by': 'system',
                        'fact_type_id': app_config['project_fact_type_id'],
                        'max_value': float(upper_bound),
                        'min_value': lower_bound,
                    })
                lower_bound = float(upper_bound) + 0.01

    logger.info('Created %r project fact type with ID %r', fact_name,
                app_config['project_fact_type_id'])



import re

from imbi.endpoints import base


class RequestHandler(base.AuthenticatedRequestHandler):

    NAME = 'reports-system-shs-history'

    SQL = re.sub(
        r'\s+', ' ', """\
        SELECT scored_on,
               ((sum(total_project_score) / sum(total_possible_score))
                    * 100::FLOAT)::NUMERIC(9,2) AS health_score
          FROM v1.namespace_kpi_history
      GROUP BY scored_on
      ORDER BY scored_on ASC""")

    async def get(self):
        result = await self.postgres_execute(self.SQL, metric_name=self.NAME)
        self.send_response(result.rows)



from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):
    NAME = 'project-secrets'
    ID_KEY = ['project_id', 'name']
    ITEM_NAME = 'project-secret'
    FIELDS = ['project_id']
    TTL = 300

    COLLECTION_SQL = (
        'SELECT s.project_id, p.slug AS project_slug, s.name, s.value,'
        '       s.created_by, s.last_modified_by'
        '  FROM v1.project_secrets AS s'
        '  JOIN v1.projects AS p ON s.project_id = p.id'
        ' WHERE s.project_id = %(project_id)s')

    async def get(self, *args, **kwargs):
        result = await self.postgres_execute(self.COLLECTION_SQL,
                                             kwargs,
                                             metric_name=f'get-{self.NAME}')
        for row in result.rows:
            row['value'] = self.application.decrypt_value(row['value'])
        self.send_response(result.rows)



from __future__ import annotations

import operator
from collections import abc

import pydantic
import semantic_version.base
from pydantic_core import core_schema


def _coerce_to_version(item: object) -> semantic_version.Version | object:
    if isinstance(item, str):
        return semantic_version.Version.coerce(item)
    return item


class VersionRange(abc.Container[semantic_version.Version]):
    spec: str
    parts: list[str]
    constraint: semantic_version.base.BaseSpec

    def __init__(self,
                 spec: str,
                 *,
                 constraint_cls: type = semantic_version.NpmSpec) -> None:
        self.spec = spec
        self.constraint = constraint_cls(spec)

    def __contains__(self, item: object) -> bool:
        return _coerce_to_version(item) in self.constraint

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.constraint})'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(spec={self.spec!r})'

    def __eq__(self, other: object) -> bool:
        if isinstance(other, VersionRange):
            return (self.spec == other.spec
                    and self.constraint == other.constraint)
        return False

    def __hash__(self) -> int:
        return hash((self.spec, self.constraint))

    @staticmethod
    def _validate(obj: str | VersionRange) -> 'VersionRange':
        if isinstance(obj, str):
            return VersionRange(obj)
        if isinstance(obj, VersionRange):
            return obj
        raise TypeError('Invalid version range')

    @classmethod
    def __get_pydantic_core_schema__(
            cls, _source_type: type,
            _handler: pydantic.GetCoreSchemaHandler) -> core_schema.CoreSchema:
        validator = core_schema.no_info_plain_validator_function(cls._validate)
        return core_schema.json_or_python_schema(
            json_schema=validator,
            python_schema=validator,
            serialization=core_schema.plain_serializer_function_ser_schema(
                operator.attrgetter('spec')),
        )

    def __get_pydantic_json_schema__(self, *args) -> dict:
        return {
            'title': 'Range of acceptable versions',
            'type': 'string',
            'externalDocs': {
                'url': 'https://www.npmjs.com/package/semver'
            },
            'minLength': 1,
        }


class ExactRange(VersionRange):
    """Exact matches only"""
    def __init__(self, spec: str, /) -> None:
        super().__init__(spec, constraint_cls=semantic_version.SimpleSpec)
        self.version = semantic_version.Version.coerce(spec)

    def __contains__(self, item: object) -> bool:
        # NB -- Version('1.2.3.1') in SimpleSpec('1.2.3') is TRUE
        return _coerce_to_version(item) == self.version


def parse_semver_range(spec: str) -> VersionRange:
    """Parse a semver range and return a matcher

    Valid "semver ranges" are defined by https://www.npmjs.com/package/semver.
    Imbi accepts only caret ranges and tilde ranges. Anything else is
    treated as an exact value match.

    """
    if spec.startswith(('~', '^')):
        return VersionRange(spec)
    return ExactRange(spec)



"""
Integration Endpoints

"""
from tornado import web

from . import (
    automations,
    github,
    gitlab,
    google,
    integrations,
    notifications,
    oauth2,
)

SLUG = r'[\w_\-%\+]+'

URLS = [
    web.url(r'^/github/auth', github.RedirectHandler, name='github-callback'),
    web.url(r'^/gitlab/auth', gitlab.RedirectHandler, name='gitlab-callback'),
    web.url(r'^/gitlab/namespaces', gitlab.UserNamespacesHandler),
    web.url(r'^/gitlab/projects', gitlab.ProjectsHandler),
    web.url(r'^/google/auth', google.RedirectHandler),
    web.url(r'^/integrations$',
            integrations.CollectionRequestHandler,
            name='integrations'),
    web.url(fr'^/integrations/(?P<name>{SLUG})$',
            integrations.RecordRequestHandler,
            name='integration'),
    web.url(fr'/integrations/(?P<integration_name>{SLUG})/automations/?',
            automations.CollectionRequestHandler,
            name='automations'),
    web.url(
        fr'/integrations/(?P<integration_name>{SLUG})'
        fr'/automations/(?P<slug>{SLUG})',
        automations.RecordRequestHandler,
        name='automation'),
    web.url(fr'/integrations/(?P<integration_name>{SLUG})/oauth2',
            oauth2.RecordRequestHandler,
            name='oauth2-management'),
] + notifications.URLS



import json
import unittest.mock
from collections import abc

import pydantic
import semantic_version

from imbi import semver


class VersionCheckingTests(unittest.TestCase):
    def verify_range(self, spec: str, in_range: abc.Iterable[str],
                     out_of_range: abc.Iterable[str]) -> None:
        ver_range = semver.parse_semver_range(spec)
        for value in in_range:
            self.assertIn(value, ver_range, value)
        for value in out_of_range:
            self.assertNotIn(value, ver_range, value)

    def test_tilde_ranges_from_documentation(self) -> None:
        # https://www.npmjs.com/package/semver#user-content-tilde-ranges-123-12-1

        # ~1.2.3 := >=1.2.3 <1.(2+1).0 := >=1.2.3 <1.3.0-0
        self.verify_range('~1.2.3', ('1.2.3', '1.2.4', '1.2.99'),
                          ('1.1', '1.2.2', '1.2.2.99', '1.2', '1.3'))

        # ~1.2 := >=1.2.0 <1.(2+1).0 := >=1.2.0 <1.3.0-0
        self.verify_range('~1.2', ('1.2', '1.2.2', '1.2.3', '1.2.99'),
                          ('1.1', '1.3.0', '1.3'))

        # ~1 := >=1.0.0 <(1+1).0.0 := >=1.0.0 <2.0.0-0
        self.verify_range('~1', ('1.2', '1.2.2', '1.2.3', '1.2.99', '1.99'),
                          ('2.0', '0.1.2', '3'))

        # ~0.2.3 := >=0.2.3 <0.(2+1).0 := >=0.2.3 <0.3.0-0
        self.verify_range('~0.2.3', ('0.2.3', '0.2.99'),
                          ('0.2.2', '0.3', '1.0'))

        # ~0.2 := >=0.2.0 <0.(2+1).0 := >=0.2.0 <0.3.0-0
        self.verify_range('~0.2', ('0.2.0', '0.2.99'), ('0.1', '0.3', '1.0'))

        # ~0 := >=0.0.0 <(0+1).0.0 := >=0.0.0 <1.0.0-0
        self.verify_range('~0', ('0.9.9', '0.1', '0'), ('1.0', '2.0'))

    def test_caret_ranges_from_documentation(self) -> None:
        # https://www.npmjs.com/package/semver#caret-ranges-123-025-004

        # ^1.2.3 := >=1.2.3 <2.0.0-0
        self.verify_range('^1.2.3', ('1.2.3', '1.2.4', '1.9.9'),
                          ('1.2.2', '2'))

        # ^0.2.3 := >=0.2.3 <0.3.0-0
        self.verify_range('^0.2.3', ('0.2.3', '0.2.4', '0.2.9'),
                          ('0.2.2', '0.3', '0', '1'))

        # ^0.0.3 := >=0.0.3 <0.0.4-0
        self.verify_range('^0.0.3', ('0.0.3', '0.0.3.1'),
                          ('0.0.2', '0.0.4', '0', '1'))

    def test_exact_ranges(self) -> None:
        self.verify_range('1.2.3', ['1.2.3'],
                          ['1.2', '1.2.2', '1.2.3.1', '1.2.4'])

    def test_caret_range_of_all_zeros(self) -> None:  # this is an edge case
        self.verify_range('^0', ('0', '0.1', '0.99.99'), ('1', '1.0'))
        self.verify_range('^0.0', ('0', '0.0.1', '0.0.99'), ('0.1', '1.0'))
        self.verify_range('^0.0.0', ('0', '0.0.0'), ('0.0.1', '0.1', '1.0'))


class VersionRangeErrorHandlingTests(unittest.TestCase):
    def test_that_empty_spec_is_rejected(self) -> None:
        with self.assertRaises(ValueError):
            semver.ExactRange('')

        with self.assertRaises(ValueError):
            semver.parse_semver_range('')

    def test_contains_works_with_str(self) -> None:
        self.assertIn('1.2.3', semver.parse_semver_range('1.2.3'))

    def test_contains_works_with_version(self) -> None:
        self.assertIn(semantic_version.Version('1.2.3'),
                      semver.parse_semver_range('1.2.3'))

    def test_contains_with_other_instance_types(self) -> None:
        ver_range = semver.parse_semver_range('1')
        self.assertNotIn(1, ver_range)
        self.assertNotIn(True, ver_range)
        self.assertNotIn(object(), ver_range)


class PydanticModelTests(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.adapter = pydantic.TypeAdapter(semver.VersionRange)
        self.value = semver.VersionRange('~1.2.3')

    def test_core_schema_validation(self) -> None:
        self.assertEqual(self.value, self.adapter.validate_json('"~1.2.3"'))
        self.assertEqual(self.value, self.adapter.validate_python(self.value))

    def test_core_schema_serialization(self) -> None:
        result = json.loads(self.adapter.serializer.to_json(self.value))
        self.assertEqual(self.value.spec, result)
        self.assertEqual(self.value.spec,
                         self.adapter.serializer.to_python(self.value))


class UtilityFunctionTests(unittest.TestCase):
    def test_equality(self) -> None:
        self.assertEqual(semver.VersionRange('1'), semver.VersionRange('1'))
        self.assertNotEqual(semver.VersionRange('1'), '1')

    def test_hashing(self) -> None:
        v1 = semver.VersionRange('1.2.3')
        v2 = semver.VersionRange('~1.2.3')
        self.assertEqual(hash(v1), hash(semver.VersionRange(v1.spec)))
        self.assertEqual(hash(v1), hash(v1))
        self.assertNotEqual(hash(v1), hash(v2))

    def test_repr(self) -> None:
        v1 = semver.VersionRange('1.2.3')
        v2 = eval(repr(v1), {'VersionRange': semver.VersionRange})
        self.assertEqual(v1, v2)

    def test_str(self) -> None:
        v = semver.VersionRange('~1.2.3')
        self.assertEqual('VersionRange(~1.2.3)', str(v))

    def test_validation(self) -> None:
        v = semver.VersionRange('~1.2.3')
        self.assertEqual(v, semver.VersionRange._validate(v))
        self.assertEqual(v, semver.VersionRange._validate('~1.2.3'))
        with self.assertRaises(TypeError):
            semver.VersionRange._validate(1.2)

    def test_json_schema(self) -> None:
        schema = pydantic.TypeAdapter(semver.VersionRange).json_schema()
        self.assertEqual('string', schema['type'])
        self.assertEqual(1, schema['minLength'])



import logging
import urllib.parse

import aioredis
import google.auth.exceptions
import google.oauth2.id_token
import sprockets.mixins.http
import yarl
from google.auth.transport import requests

import imbi.user
from imbi import errors, oauth2
from imbi.endpoints import base

LOGGER = logging.getLogger(__name__)


class RedirectHandler(sprockets.mixins.http.HTTPClientMixin,
                      base.RequestHandler):
    integration: oauth2.OAuth2Integration

    NAME = 'google-redirect'

    @property
    def _redis(self) -> aioredis.Redis:
        return self.application.session_redis

    async def prepare(self) -> None:
        await super().prepare()
        if not self._finished:
            self.integration = await oauth2.OAuth2Integration.by_name(
                self.application, self.integration_name)
            if not self.integration:
                raise errors.IntegrationNotFound(self.integration_name)

    async def get(self):
        state = self.get_query_argument('state', default=None)
        if not state:
            raise errors.BadRequest('No state in request payload')
        redis_state = await self._redis.get(state)
        if not redis_state:
            raise errors.Forbidden('Returned state CRSF token not found')
        await self._redis.delete(state)

        code = self.get_query_argument('code', default=None)
        if not code:
            raise errors.BadRequest('No code in request payload')

        (id_token, access_token,
         refresh_token) = await self.exchange_code_for_tokens(code)

        user = await self.sync_user(id_token)
        await self.save_session(user)
        await self.integration.upsert_user_tokens(user.username,
                                                  user.external_id,
                                                  access_token, refresh_token,
                                                  id_token)
        target = yarl.URL(self.request.full_url()).with_path('/ui/')
        self.redirect(str(target))

    async def save_session(self, user):
        self.session.user = user
        await self.session.save()

    async def sync_user(self, id_token) -> imbi.user.User:
        try:
            id_info = google.oauth2.id_token.verify_oauth2_token(
                id_token, requests.Request(), self.integration.client_id)
        except google.auth.exceptions.GoogleAuthError as e:
            raise errors.BadRequest('Token issuer is invalid: %s', e)
        except ValueError:
            raise errors.BadRequest('Invalid token')
        if id_info['aud'] != self.integration.client_id:
            raise errors.BadRequest('Invalid token: client ID does not match')
        if id_info['hd'] not in self.settings['google']['valid_domains']:
            raise errors.Forbidden('Invalid Google domain sign in')
        user_id = id_info['sub']
        user_email = id_info['email']
        display_name = id_info['name']
        username = user_email.split('@')[0]
        imbi_user = imbi.user.User(self.application,
                                   username=username,
                                   google_user=True,
                                   external_id=user_id,
                                   display_name=display_name,
                                   email_address=user_email)
        await imbi_user.refresh()
        return imbi_user

    async def exchange_code_for_tokens(self, code):
        body = urllib.parse.urlencode({
            'client_id': self.integration.client_id,
            'client_secret': self.integration.client_secret,
            'grant_type': 'authorization_code',
            'redirect_uri': self.integration.callback_url,
            'code': code,
        })
        response = await self.http_fetch(
            str(self.integration.token_endpoint),
            method='POST',
            body=body,
            content_type='application/x-www-form-urlencoded')
        if not response.ok:
            self.logger.error('failed to exchange auth code for token: %s %s',
                              response.body['error'],
                              response.body['error_description'])
            raise errors.InternalServerError(
                'failed exchange auth code for token: %s',
                response.code,
                title='Google authorization failure',
                instance={
                    'error': response.body['error'],
                    'error_description': response.body['error_description'],
                })
        return (response.body.get('id_token'),
                response.body.get('access_token'),
                response.body.get('refresh_token'))

    @property
    def integration_name(self):
        return self.settings['google']['integration_name']



from __future__ import annotations

import base64
import contextlib
import io
import os
import tempfile
import unittest.mock
import uuid

import yaml

from imbi import server


class ConfigurationTestCase(unittest.TestCase):
    temp_file: io.FileIO

    def setUp(self) -> None:
        super().setUp()
        self._exit_stack = contextlib.ExitStack()
        self.addCleanup(self._exit_stack.close)
        self.mock_stderr = self.patch_object(server.sys, 'stderr')
        self.temp_file = self._exit_stack.enter_context(
            tempfile.NamedTemporaryFile(mode='w+t', encoding='utf-8'))
        self.config = {'http': {'canonical_server_name': 'server.example.com'}}

    def patch_object(
            self, target: object, attribute: str,
            **kwargs) -> unittest.mock.MagicMock | unittest.mock.AsyncMock:
        return self._exit_stack.enter_context(
            unittest.mock.patch.object(target, attribute, **kwargs))


class LoadConfigurationTests(ConfigurationTestCase):
    def test_missing_configuration_file(self):
        with self.assertRaises(SystemExit):
            server.load_configuration(str(uuid.uuid4()), False)
        self.mock_stderr.write.assert_called()

    def test_yaml_load_failure(self):
        self.temp_file.write('\x00\x01\x02\x03')
        self.temp_file.flush()
        with self.assertRaises(SystemExit):
            server.load_configuration(self.temp_file.name, False)
        self.mock_stderr.write.assert_called()

    def test_bad_yaml_doc(self):
        self.temp_file.write('<config/>')
        self.temp_file.flush()
        with self.assertRaises(SystemExit):
            server.load_configuration(self.temp_file.name, False)
        self.mock_stderr.write.assert_called()

    def test_encryption_key_encoding(self):
        secret = os.urandom(32)
        self.config['encryption_key'] = base64.b64encode(secret).decode()
        yaml.dump(self.config, self.temp_file)
        config, _ = server.load_configuration(self.temp_file.name, False)
        self.assertEqual(secret, config['encryption_key'])

        self.temp_file.seek(0)
        self.config['encryption_key'] = 'not valid base64'
        yaml.dump(self.config, self.temp_file)
        config, _ = server.load_configuration(self.temp_file.name, False)
        self.assertEqual('not valid base64'.encode(), config['encryption_key'])


class SentryConfigurationTests(ConfigurationTestCase):
    def test_that_default_is_enabled(self):
        yaml.dump(self.config, self.temp_file)
        config, _ = server.load_configuration(self.temp_file.name, False)
        self.assertTrue(config['automations']['sentry']['enabled'])

    def test_that_sentry_can_be_disabled(self):
        self.config['automations'] = {'sentry': {'enabled': False}}
        yaml.dump(self.config, self.temp_file)
        loaded, _ = server.load_configuration(self.temp_file.name, False)
        self.assertFalse(loaded['automations']['sentry']['enabled'])

    def test_that_sentry_url_has_sensible_default(self):
        yaml.dump(self.config, self.temp_file)
        config, _ = server.load_configuration(self.temp_file.name, False)
        self.assertEqual('https://sentry.io/',
                         config['automations']['sentry']['url'])


class GitLabConfiguratTests(ConfigurationTestCase):
    def test_that_default_is_disabled(self):
        yaml.dump(self.config, self.temp_file)
        config, _ = server.load_configuration(self.temp_file.name, False)
        self.assertFalse(config['automations']['gitlab']['enabled'])

    def test_that_gitlab_is_enabled_when_fully_configured(self):
        self.config['automations'] = {'gitlab': {}}
        yaml.dump(self.config, self.temp_file)
        loaded, _ = server.load_configuration(self.temp_file.name, False)
        self.assertFalse(loaded['automations']['gitlab']['enabled'])
        self.temp_file.seek(0)

        self.config['automations']['gitlab']['project_link_type_id'] = 1
        yaml.dump(self.config, self.temp_file)
        loaded, _ = server.load_configuration(self.temp_file.name, False)
        self.assertTrue(loaded['automations']['gitlab']['enabled'])



from __future__ import annotations

import dataclasses
import datetime
import inspect
import logging
import types

import graphlib
import sprockets_postgres  # type: ignore[import-untyped]
import typing_extensions as typing

from imbi import errors

if typing.TYPE_CHECKING:  # pragma: nocover
    import imbi.app
    import imbi.models
    import imbi.user

CompensatingAction: typing.TypeAlias = typing.Callable[
    ['AutomationContext', BaseException], typing.Union[typing.Awaitable[None],
                                                       None]]


class QueryFunction(typing.Protocol):
    async def __call__(
        self,
        sql: str,
        parameters: sprockets_postgres.QueryParameters = None,
        metric_name: str = '',
        *,
        timeout: sprockets_postgres.Timeout = None
    ) -> sprockets_postgres.QueryResult:  # pragma: nocover
        ...


async def do_nothing(context: AutomationContext,
                     automation: imbi.models.Automation, *_args: object,
                     **_kwargs: object) -> None:
    context.note_progress('default action running for %s', automation.slug)


class AutomationFailedError(Exception):
    def __init__(self, automation: imbi.models.Automation,
                 error: Exception) -> None:
        super().__init__(f'Automation {automation.slug} failed: {error}')
        self.automation = automation
        self.error = error


@dataclasses.dataclass
class AutomationNote:
    when: datetime.datetime
    integration_name: str | None
    automation_slug: str | None
    message: str


class AutomationContext:
    """Run automations with automated cleanup & injected context

    You may want to use the `run_automations` function instead of
    directly creating a context instance.

    The automation context provides a convenient way to run automation
    *actions*. Each *action* is a callable that accepts the context
    instance as it's first parameter. Additional parameters are passed
    via `args` & `kwargs`.

    The context also has a stack of *compensating callbacks* that will
    be invoked if an exception is thrown during the processing of an
    action. The callback stack is *only* invoked when an exception occurs
    and each callback will be invoked exactly once. It is passed the
    context instance and the exception instance that was caught.

    The context has a number of attributes exported for use inside an
    action.

    - *current_automation* is set to the slug of the executing automation
      or ``None`` if no automation is active.
    - *current_integration* is set to the name of the executing automations
      integration or ``None`` if no automation is active.
    - *run_query* is a handle to the sprockets_postgres postgres_execute
      method. Use this for database access.
    - *user* is a `imbi.models.User` instance for the acting user. You
      can use this to retrieve user-specific tokens for external apps
      such as gitlab.

    """
    def __init__(self, application: imbi.app.Application, user: imbi.user.User,
                 query: QueryFunction) -> None:
        self.application = application
        self.logger = logging.getLogger(__name__).getChild('AutomationContext')
        self.notes: list[AutomationNote] = []
        self.user = user
        self.run_query = query

        # these are managed inside `run_automation`
        self._current_automation: str | None = None
        self._current_integration: str | None = None

        self._cleanups: list[CompensatingAction] = []
        self._memory: dict[object, object] = {}

    @property
    def current_automation(self) -> str | None:
        return self._current_automation

    @current_automation.setter
    def current_automation(self, a: imbi.models.Automation | None) -> None:
        if a is None:
            self._current_automation = None
            self._current_integration = None
        else:
            self._current_automation = a.slug
            self._current_integration = a.integration_name

    @property
    def current_integration(self) -> str | None:
        return self._current_integration

    def note_progress(self, message_format: str, *args: object) -> None:
        """Add a note to the list of notes

        These could be made available in an API response so don't
        log anything sensitive.
        """
        prefix = ''
        if self.current_automation:
            prefix = f'{self.current_integration}/{self.current_automation} '
        self.logger.info(prefix + message_format, *args)
        self.notes.append(
            AutomationNote(when=datetime.datetime.now(datetime.timezone.utc),
                           integration_name=self.current_integration,
                           automation_slug=self.current_automation,
                           message=message_format %
                           args if args else message_format))

    async def run_automation(self, automation: imbi.models.Automation,
                             *args: object, **kwargs: object) -> None:
        self.logger.debug('running automation %s = %r', automation.slug,
                          automation.callable)
        try:
            self.current_automation = automation
            result = automation.callable(self, automation, *args, **kwargs)
            if inspect.isawaitable(result):
                await result
        except Exception as error:
            self.logger.exception(
                'automation %s failed with %r, running %s cleanups',
                automation.slug, error, len(self._cleanups))
            await self._run_cleanups(error)
            raise error
        finally:
            self.current_automation = None

    def add_callback(self, compensating_action: CompensatingAction) -> None:
        """Add a cleanup action"""
        self._cleanups.append(compensating_action)

    # Implement MutableMapping interface for state storage
    def __getitem__(self, key: object) -> object:
        return self._memory[key]

    def __delitem__(self, key: object) -> None:
        del self._memory[key]

    def __setitem__(self, key: object, value: object) -> None:
        self._memory[key] = value

    def __contains__(self, key: object) -> bool:
        return key in self._memory

    # Implement an async context manager that runs cleanups if an
    # exception is raised
    async def __aenter__(self) -> typing.Self:
        return self

    async def __aexit__(self, _exc_type: typing.Type[BaseException] | None,
                        exc_val: BaseException | None,
                        _exc_tb: types.TracebackType | None) -> None:
        if exc_val is not None:
            await self._run_cleanups(exc_val)

    async def _run_cleanups(self, exc: BaseException) -> None:
        cleanups, self._cleanups = self._cleanups, []
        for cleanup in reversed(cleanups):
            try:
                maybe_coro = cleanup(self, exc)
                if inspect.isawaitable(maybe_coro):
                    await maybe_coro
            except Exception as error:
                self.logger.exception('cleanup %r failed with %s', cleanup,
                                      error)


async def run_automations(
        automations: typing.Iterable[imbi.models.Automation],
        *args: object,
        application: imbi.app.Application,
        user: imbi.user.User,
        query_executor: sprockets_postgres.RequestHandlerMixin,
        addt_callbacks: typing.Iterable[CompensatingAction] | None = None,
        **kwargs: object) -> None:
    """Run a list of operations in a AutomationContext instance

    `args` and `kwargs` are passed as-is to **every** automation
    callable. If an automation fails, then a AutomationFailedError
    will be raised instead of whatever the automation raised.
    """
    last_automation = None
    try:
        async with AutomationContext(application=application,
                                     query=query_executor.postgres_execute,
                                     user=user) as context:
            for cb in addt_callbacks or []:
                context.add_callback(cb)
            for automation in automations:
                last_automation = automation
                await context.run_automation(automation, *args, **kwargs)
                # TODO dave-shawley:
                #   it would be nice to optionally refresh *args and
                #   **kwargs after running an automation. Consider
                #   project-creation automations that add identifiers
                #   and pass the project as args[0]... Currently they
                #   have to query for identifiers which requires that
                #   they know the name of the other integration :/
    except errors.ApplicationError:  # this is meant for the end user
        raise
    except Exception as error:
        # only note which automation failed if we ran one
        # AutomationContext.run_automation logs an "action failed" message,
        # so we don't have to
        if last_automation is not None:
            raise AutomationFailedError(last_automation, error) from None
        raise


def query_runner(
    metric_name: str,
    query: str,
    *args: sprockets_postgres.QueryParameters | str,
    **kwargs: sprockets_postgres.Timeout | str,
) -> CompensatingAction:
    """Returns a CompensatingAction that calls context.run_query()

    Use this to create compensating actions that run a simple
    SQL query. If you need to do something more in-depth, then
    write freestanding function.
    """
    kwargs['metric_name'] = metric_name

    async def runner(context: AutomationContext, *_: object) -> None:
        context.note_progress('running query %r with args %r', metric_name,
                              args)
        await context.run_query(query, *args, **kwargs)

    return runner


async def retrieve_automations(
        application: imbi.app.Application, slugs: typing.Iterable[str],
        project_type_id: int) -> list[imbi.models.Automation]:
    """Retrieve a list of automations by slug and validate them

    If a slug doesn't exist in the database or any automation is
    not applicable to the project type, then an exception is raised.

    """
    from imbi import models  # prevent circular import
    automation_instances = [
        await models.automation(slug, application) for slug in slugs
    ]
    required_automations = set()
    for a in automation_instances:
        required_automations.update(a.depends_on)
    required_automations.difference_update(
        {a.slug
         for a in automation_instances})
    error = {
        'nonexistent_automations': [
            name for name, matched in zip(slugs, automation_instances)
            if matched is None
        ],
        'invalid_automations': [
            a.name for a in automation_instances
            if a is not None and project_type_id not in a.applies_to_ids
        ],
        'missing_required_automations': list(required_automations),
    }
    error = {label: value for label, value in error.items() if value}
    if error:
        raise errors.BadRequest('Invalid automation request', **error)
    return automation_instances


def sort_automations(
    automations: typing.Sequence[imbi.models.Automation]
) -> typing.Sequence[imbi.models.Automation]:
    if not automations:
        return automations

    lookup = {a.slug: a for a in automations}
    dag = graphlib.TopologicalSorter()
    for automation in automations:
        dag.add(automation.slug, *automation.depends_on)
    return [lookup[a] for a in dag.static_order()]



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = 'name'
    FIELDS = ['name', 'group_type', 'external_id', 'permissions']
    DEFAULTS = {'group_type': 'internal', 'permissions': []}

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT "name", created_at, created_by,
               last_modified_at, last_modified_by,
               group_type, external_id, permissions
          FROM v1.groups
         WHERE "name"=%(name)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'groups'
    ITEM_NAME = 'group'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT "name", group_type, external_id, permissions
          FROM v1.groups
         ORDER BY "name" ASC""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.groups
                    ("name", created_by, group_type, external_id, permissions)
             VALUES (%(name)s, %(username)s, %(group_type)s, %(external_id)s,
                     %(permissions)s)
          RETURNING "name";""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'group'

    DELETE_SQL = 'DELETE FROM v1.groups WHERE "name"=%(name)s;'

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.groups
           SET "name"=%(name)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               group_type=%(group_type)s,
               external_id=%(external_id)s,
               permissions=%(permissions)s
         WHERE "name"=%(current_name)s""")



from __future__ import annotations

import http
import typing

import jsonpatch
import pydantic

from imbi import errors, postgres
from imbi.endpoints import base


class OAuth2CreationRequest(pydantic.BaseModel):
    authorization_endpoint: pydantic.HttpUrl
    token_endpoint: pydantic.HttpUrl
    revoke_endpoint: typing.Union[pydantic.HttpUrl, None]
    callback_url: typing.Union[pydantic.HttpUrl, None]
    client_id: str
    client_secret: typing.Union[str, None]
    use_pkce: bool


class OAuth2Details(pydantic.BaseModel):
    authorization_endpoint: pydantic.HttpUrl
    token_endpoint: pydantic.HttpUrl
    revoke_endpoint: typing.Union[pydantic.HttpUrl, None] = None
    callback_url: typing.Union[pydantic.HttpUrl, None]
    client_id: str
    client_secret: typing.Union[str, None]
    use_pkce: bool

    @pydantic.field_serializer('client_secret')
    def hide_secret(self, v) -> str:
        return None if v is None else '********'


class RecordRequestHandler(base.PydanticHandlerMixin,
                           base.AuthenticatedRequestHandler):
    NAME = 'oauth2-management'

    async def _get_oauth2_details(
            self, integration_name: str) -> OAuth2Details | None:
        result = await self.postgres_execute(
            'SELECT authorization_endpoint, token_endpoint, revoke_endpoint,'
            '       callback_url, client_id, client_secret,'
            '       public_client AS use_pkce'
            '  FROM v1.oauth2_integrations'
            ' WHERE name = %(integration_name)s',
            {'integration_name': integration_name})
        return None if not result else OAuth2Details.model_validate(result.row)

    @base.require_permission('admin')
    async def get(self, integration_name: str) -> None:
        details = await self._get_oauth2_details(integration_name)
        if not details:
            raise errors.ItemNotFound(
                'Integration %r does not support OAuth2 connections',
                integration_name)
        self.send_response(details)

    @base.require_permission('admin')
    async def post(self, integration_name: str) -> None:
        request = self.parse_request_body_as(OAuth2CreationRequest)

        if await self._get_oauth2_details(integration_name):
            raise errors.ApplicationError(
                http.HTTPStatus.CONFLICT,
                'Integration %r already has OAuth2 connection details',
                integration_name)

        client_secret = None
        if request.client_secret:
            client_secret = request.client_secret

        await self.postgres_execute(
            'INSERT INTO v1.oauth2_integrations ('
            '            name, api_endpoint, callback_url,'
            '            authorization_endpoint, token_endpoint,'
            '            revoke_endpoint, client_id, client_secret,'
            '            public_client)'
            '     SELECT i.name, i.api_endpoint, %(callback_url)s,'
            '            %(authorization_endpoint)s, %(token_endpoint)s,'
            '            %(revoke_endpoint)s, %(client_id)s,'
            '            %(client_secret)s, %(public_client)s'
            '       FROM v1.integrations AS i'
            '      WHERE i.name = %(integration_name)s', {
                'integration_name': integration_name,
                'callback_url': (str(request.callback_url)
                                 if request.callback_url else None),
                'authorization_endpoint': str(request.authorization_endpoint),
                'token_endpoint': str(request.token_endpoint),
                'revoke_endpoint': (str(request.revoke_endpoint)
                                    if request.revoke_endpoint else None),
                'client_id': request.client_id,
                'client_secret': client_secret,
                'public_client': request.use_pkce,
            })
        self.send_response(
            OAuth2Details.model_validate({
                'authorization_endpoint': request.authorization_endpoint,
                'token_endpoint': request.token_endpoint,
                'revoke_endpoint': request.revoke_endpoint,
                'callback_url': request.callback_url,
                'client_id': request.client_id,
                'client_secret': request.client_secret,
                'use_pkce': request.use_pkce,
            }))

    @base.require_permission('admin')
    async def delete(self, integration_name: str) -> None:
        result = await self.postgres_execute(
            'DELETE FROM v1.oauth2_integrations'
            '      WHERE name = %(integration_name)s',
            {'integration_name': integration_name},
            metric_name='delete-oauth2-details')
        if not result:
            raise errors.ItemNotFound(
                'Integration %r does not support OAuth2 connections',
                integration_name)
        self.set_status(http.HTTPStatus.NO_CONTENT, reason='Item Deleted')

    @base.require_permission('admin')
    async def patch(self, integration_name: str) -> None:
        try:
            patch = jsonpatch.JsonPatch(self.get_request_body())
        except (jsonpatch.JsonPatchException,
                jsonpatch.JsonPointerException) as error:
            raise errors.BadJsonPatch(error)

        details = await self._get_oauth2_details(integration_name)
        if not details:
            raise errors.ItemNotFound(
                'Integration %r does not support OAuth2 connections',
                integration_name)

        original = details.model_dump(mode='json')
        if original['client_secret']:  # enable `test` on real client secret
            original['client_secret'] = details.client_secret
        # limit operations to protect secret
        patch.operations = {
            name: operation
            for name, operation in patch.operations.items()
            if name not in {'copy', 'move'}
        }

        updated = patch.apply(original)
        try:
            OAuth2Details.model_validate(updated)
        except pydantic.ValidationError as error:
            raise errors.PydanticValidationError(
                error, 'Failed to validate new OAuth2 details') from None

        changed = {
            column
            for column in OAuth2Details.model_fields.keys()
            if updated[column] != original[column]
        }
        if not changed:
            self.set_status(http.HTTPStatus.NOT_MODIFIED)
            return

        if updated['client_secret']:
            updated['client_secret'] = updated['client_secret']
        async with self.application.postgres_connector(
                self.on_postgres_error, self.on_postgres_timing) as conn:
            # handle mismatch between representation and DB
            original['public_client'] = original.pop('use_pkce')
            updated['public_client'] = updated.pop('use_pkce')
            if 'use_pkce' in changed:
                changed.remove('use_pkce')
                changed.add('public_client')

            await postgres.update_entity(conn,
                                         'v1',
                                         'oauth2_integrations',
                                         original,
                                         updated,
                                         changed,
                                         id_column='name',
                                         id_value=integration_name)

        self.send_response(await self._get_oauth2_details(integration_name))



import unittest

import tornado.httputil

from imbi import errors


class DefaultFunctionalityTests(unittest.TestCase):
    def test_that_error_url_can_be_configured(self):
        saved_error_url = errors.ERROR_URL
        try:
            errors.set_canonical_server('server.example.com')
            err = errors.ApplicationError(500, 'error-fragment', '')
            self.assertEqual('https://server.example.com/#error-fragment',
                             err.document['type'])
        finally:
            errors.ERROR_URL = saved_error_url

    def test_that_reason_is_set_from_status_code(self):
        err = errors.ApplicationError(500, 'fragment', '')
        self.assertEqual(tornado.httputil.responses[500], err.reason)

    def test_that_reason_is_title_if_set(self):
        err = errors.ApplicationError(500, 'fragment', '', title='title')
        self.assertEqual('title'.title(), err.reason)

    def test_that_unknown_status_codes_are_handled(self):
        err = errors.ApplicationError(1, 'fragment', '')
        self.assertEqual('Unknown Status Code', err.reason)

        err = errors.ApplicationError(600, 'fragment', '')
        self.assertEqual('Unknown Status Code', err.reason)

    def test_that_type_can_be_overridden(self):
        err = errors.ApplicationError(
            500,
            'error-fragment',
            '',
            type='https://example.com/troubleshooting')
        self.assertEqual('https://example.com/troubleshooting',
                         err.document['type'])

    def test_that_title_defaults_to_reason(self):
        err = errors.ApplicationError(500, 'fragment', '')
        self.assertEqual(tornado.httputil.responses[500],
                         err.document['title'])

    def test_that_detail_defaults_to_formatted_log_message(self):
        err = errors.ApplicationError(500, 'fragment', '1+1=%s', 1 + 1)
        self.assertEqual('1+1=2', err.document['detail'])

    def test_with_missing_log_args(self):
        err = errors.ApplicationError(500, 'fragment', '%s')
        self.assertEqual('%s', err.document['detail'])

    def test_with_unused_log_args(self):
        err = errors.ApplicationError(500, 'fragment', 'No args', 'arg')
        self.assertEqual('No args', err.document['detail'])


class SpecificErrorBehaviorTests(unittest.TestCase):
    def test_item_not_found_default_title(self):
        err = errors.ItemNotFound()
        self.assertEqual('Item not found', err.document['title'])

    def test_item_not_found_log_message(self):
        err = errors.ItemNotFound()
        self.assertEqual('Item not found', err.log_message)

    def test_method_not_allowed_log_message(self):
        err = errors.MethodNotAllowed('post')
        self.assertEqual('POST is not a supported HTTP method',
                         err.document['detail'])

    def test_unsupported_media_type_log_message(self):
        err = errors.UnsupportedMediaType('application/xml')
        self.assertEqual('application/xml is not a supported media type',
                         err.document['detail'])

    def test_database_error_defaults(self):
        err = errors.DatabaseError()
        self.assertEqual('Database Error', err.document['title'])
        self.assertEqual('Database Error', err.reason)
        self.assertEqual('Database failure', err.log_message)

    def test_database_error_with_exception(self):
        failure = RuntimeError('whatever')
        err = errors.DatabaseError(error=failure)
        self.assertEqual('Database Error', err.document['title'])
        self.assertEqual('Database Error', err.reason)
        self.assertEqual('Database failure: %s', err.log_message)
        self.assertEqual((failure, ), err.args)

    def test_database_error_with_explicit_title(self):
        err = errors.DatabaseError(title='No rows returned')
        self.assertEqual('No rows returned', err.document['title'])
        self.assertEqual('No rows returned'.title(), err.reason)
        self.assertEqual('Database failure', err.log_message)



"""
Tornado TemplateLoader that pulls data from a Python package installation

"""
import logging
import pkgutil
from os import path

from tornado import template

LOGGER = logging.getLogger(__name__)


class TemplateLoader(template.BaseLoader):
    """A template loader that loads from Python package data."""
    def __init__(self, debug=False, **kwargs):
        """Create a new instance of the loader, respecting the debug flag
        so that when set, templates are not cached, since changing templates
        does not trigger Tornado to restart.

        :param bool debug: Tornado's debug setting
        :param dict kwargs: Optional kwargs to pass through

        """
        super(TemplateLoader, self).__init__(**kwargs)
        self.debug = debug
        self.root = 'templates'

    def load(self, name, parent_path=None):
        """Loads a template, caching results in a local dict.

        :param str name: The name of the template
        :param str parent_path: The parent path for the template
        :rtype: str

        """
        name = self.resolve_path(name, parent_path=parent_path)
        with self.lock:
            if self.debug or name not in self.templates:
                self.templates[name] = self._create_template(name)
        return self.templates[name]

    def resolve_path(self, name, parent_path=None):
        """Resolve the Tornado template relative path.

        :param str name: The name of the template
        :param str parent_path: The parent path for the template
        :rtype: str

        """
        if parent_path and not parent_path.startswith('<') and \
            not parent_path.startswith('/') and \
                not name.startswith('/'):
            current_path = path.join(self.root, parent_path)
            file_dir = path.dirname(current_path)
            relative_path = path.join(file_dir, name)
            if relative_path.startswith(self.root):
                name = relative_path[len(self.root) + 1:]
        return path.normpath(name)

    def _create_template(self, name):
        """Loads the template from package data.

        :param str name: The relative path to the template file in the package
        :rtype: tornado.template.Template

        """
        LOGGER.debug('Loading %s', path.join(self.root, name))
        return template.Template(pkgutil.get_data(
            __name__.split('.')[0], path.join(self.root, name)),
                                 name=name,
                                 loader=self)



import base64
import dataclasses
import typing

import sprockets.mixins.http
import yarl

from imbi import errors, oauth2, user
from imbi.endpoints import base


@dataclasses.dataclass
class GitHubToken:
    access_token: str
    refresh_token: str


class RedirectHandler(sprockets.mixins.http.HTTPClientMixin,
                      base.RequestHandler):
    integration: 'oauth2.OAuth2Integration'

    NAME = 'github-redirect'

    async def prepare(self) -> None:
        await super().prepare()
        if not self._finished:
            self.integration = await oauth2.OAuth2Integration.by_name(
                self.application, 'github')
            if not self.integration:
                raise errors.IntegrationNotFound('github')

    async def get(self):
        auth_code = self.get_query_argument('code')
        state = base64.b64decode(self.get_query_argument('state'),
                                 b'-_').decode('utf-8')
        username, target = state.rstrip('?').split(':')
        token = await self.exchange_code_for_token(auth_code)
        try:
            user_id, email = await self.fetch_github_user(token)
            imbi_user = user.User(self.application, username=username)
            await imbi_user.refresh()
            if imbi_user.email_address != email:
                raise errors.Forbidden(
                    'mismatched user email: expected %r received %r',
                    imbi_user.email_address,
                    email,
                    title='GitHub authorization failure',
                    detail='unexpected email address {} for user {}'.format(
                        email, imbi_user.username))
            await self.integration.upsert_user_tokens(imbi_user.username,
                                                      str(user_id),
                                                      token.access_token,
                                                      token.refresh_token)
            await imbi_user.update_last_seen_at()
        except Exception:
            await self.revoke_github_token(token, username)
            raise

        target = yarl.URL(self.request.full_url()).with_path(target or '/ui/')
        self.redirect(str(target))

    async def exchange_code_for_token(self, code) -> GitHubToken:
        body = {
            'client_id': self.integration.client_id,
            'client_secret': self.integration.client_secret,
            'grant_type': 'authorization_code',
            'redirect_uri': str(self.integration.callback_url),
            'code': code,
        }
        response = await self.http_fetch(
            str(self.integration.token_endpoint),
            method='POST',
            body=body,
            content_type=sprockets.mixins.http.CONTENT_TYPE_JSON,
            request_headers={
                'X-GitHub-Api-Version': '2022-11-28',
                'Accept': str(sprockets.mixins.http.CONTENT_TYPE_JSON)
            })
        if not response.ok:
            self.logger.error(
                'failed to exchange auth code for token: (%s) %s',
                response.code, response.body)
            raise errors.InternalServerError(
                'failed exchange auth code for token: %s',
                response.code,
                title='GitHub authorization failure',
                instance={'error': response.body},
            )
        self.logger.debug('response_body %r', response.body)
        return GitHubToken(access_token=response.body['access_token'],
                           refresh_token=response.body['refresh_token'])

    async def fetch_github_user(self, token: GitHubToken) \
            -> typing.Tuple[int, str]:
        response = await self.http_fetch(
            str(self.integration.api_endpoint / 'user'),
            request_headers={
                'X-GitHub-Api-Version': '2022-11-28',
                'Accept': str(sprockets.mixins.http.CONTENT_TYPE_JSON),
                'Authorization': f'Bearer {token.access_token}'
            })
        if response.ok:
            return (response.body['id'], response.body['email'])
        raise errors.InternalServerError(
            'failed to retrieve GitHub user from access token',
            title='GitHub user lookup failure')

    async def revoke_github_token(self, token: GitHubToken, username: str):
        credentials = []
        token_types = []
        if token.access_token:
            credentials.append(token.access_token)
            token_types.append('access_token')
        if token.refresh_token:
            credentials.append(token.refresh_token)
            token_types.append('refresh_token')

        if not credentials:
            return

        body = {'credentials': credentials}
        response = await self.http_fetch(
            str(self.integration.api_endpoint / 'credentials' / 'revoke'),
            method='POST',
            body=body,
            content_type=sprockets.mixins.http.CONTENT_TYPE_JSON,
            request_headers={
                'X-GitHub-Api-Version': '2022-11-28',
                'Accept': str(sprockets.mixins.http.CONTENT_TYPE_JSON),
            },
        )

        if not response.ok:
            self.logger.warning(
                'failed to revoke GitHub credentials for user %s: %s '
                '(tokens: %s)', username, response.code,
                ', '.join(token_types))



import dataclasses
import typing

from imbi import automations, errors, models
from imbi.clients import sonarqube


@dataclasses.dataclass
class SonarQubeContext:
    integration_name: str
    project_info: sonarqube.ProjectInfo


async def create_project(
    context: automations.AutomationContext,
    automation: models.Automation,
    project: models.Project,
) -> None:
    """Create a SonarQube project for `project`

    1. creates the SonarQube project with project key formed from
       project.namespace.slug and project.slug
    2. saves the project key as a SonarQube identifier
    3. creates a SonarQube dashboard link if the `project_link_type_id`
       is configured in self.application.settings
    4. configure the SonarQube PR decoration if the project has a
       gitlab external ID
    """
    client = await sonarqube.create_client(context.application,
                                           automation.integration_name)
    sonar_project = await client.create_project(project)
    context[create_project] = SonarQubeContext(
        integration_name=automation.integration_name,
        project_info=sonar_project)
    context.add_callback(_delete_project)
    context.note_progress('created SonarQube project %s for Imbi project %s',
                          sonar_project.key, project.id)

    # NB -- project is deleted upon failure and the delete cascades to
    # other tables so there is no need to clean up here

    await context.run_query(
        'INSERT INTO v1.project_identifiers (external_id, integration_name,'
        '                                    project_id, created_by)'
        '     VALUES (%(project_key)s, %(integration_name)s, %(project_id)s,'
        '             %(user)s)', {
            'project_key': sonar_project.key,
            'integration_name': automation.integration_name,
            'project_id': project.id,
            'user': context.user.username,
        }, 'insert-project-identifiers')
    context.note_progress(
        'registered SonarQube identifier %s for Imbi project %s',
        sonar_project.key, project.id)

    settings = context.application.settings['automations']['sonarqube']
    if (link_type_id := settings.get('project_link_type_id')) is not None:
        await context.run_query(
            'INSERT INTO v1.project_links (project_id, link_type_id,'
            '                              created_by, url)'
            '     VALUES (%(project_id)s, %(link_type_id)s, %(user)s,'
            '             %(dashboard_url)s)', {
                'project_id': project.id,
                'link_type_id': link_type_id,
                'user': context.user.username,
                'dashboard_url': str(sonar_project.dashboard_url)
            }, 'insert-project-links')
        context.note_progress('created SonarQube link %s for Imbi project %s',
                              sonar_project.dashboard_url, project.id)

    identifiers = await models.project_identifiers(project.id,
                                                   context.application)
    for identifier in identifiers:
        if identifier.integration_name == 'gitlab':  # TODO ... sigh
            context.note_progress('identifier: %s -> %s',
                                  identifier.integration_name,
                                  identifier.external_id)
            await client.enable_pr_decoration(sonar_project,
                                              int(identifier.external_id))
            break


async def _delete_project(context: automations.AutomationContext,
                          _error: BaseException) -> None:
    try:
        sonar_info = typing.cast(SonarQubeContext, context[create_project])
    except KeyError:
        return

    try:
        client = await sonarqube.create_client(context.application,
                                               sonar_info.integration_name)
    except errors.ClientUnavailableError:
        pass
    else:
        context.note_progress('removing sonarqube project s due to error',
                              sonar_info.project_info.key)
        await client.remove_project(sonar_info.project_info)



import uuid

import aioredis
import yarl

from imbi import errors, oauth2
from imbi.endpoints import base


class LoginRequestHandler(base.RequestHandler):

    NAME = 'ui-login'

    async def post(self, *args, **kwargs):
        body = self.get_request_body()
        if not await self.session.authenticate(body.get('username'),
                                               body.get('password')):
            self.logger.debug('Session failed to authenticate')
            self.set_status(401)
            self.send_response({'message': 'Authentication Failure'})
            return
        await self.session.save()
        self.set_status(200)
        self.send_response(self.session.user.as_dict())


class GoogleLoginRequestHandler(base.RequestHandler):
    integration: oauth2.OAuth2Integration

    NAME = 'ui-login-google'

    async def prepare(self) -> None:
        await super().prepare()
        if not self._finished:
            self.integration = await oauth2.OAuth2Integration.by_name(
                self.application, self.integration_name)
            if not self.integration:
                raise errors.IntegrationNotFound(self.integration_name)

    async def get(self, *args, **kwargs):
        state = str(uuid.uuid4())
        await self._redis.set(state, 'google-login-crsf', expire=86400)
        target = yarl.URL(self.integration.authorization_endpoint).with_query({
            'client_id': str(self.integration.client_id),
            'redirect_uri': str(self.integration.callback_url),
            'state': state,
            'response_type': 'code',
            'scope': ' '.join([
                'openid', 'https://www.googleapis.com/auth/userinfo.email',
                'https://www.googleapis.com/auth/userinfo.profile'
            ]),
            'access_type': 'offline',
            'include_granted_scopes': 'true',
            'prompt': 'select_account',
        })
        self.redirect(str(target))

    @property
    def _redis(self) -> aioredis.Redis:
        return self.application.session_redis

    @property
    def integration_name(self):
        return self.settings['google']['integration_name']


class LogoutRequestHandler(base.RequestHandler):

    NAME = 'ui-logout'

    async def get(self, *args, **kwargs):
        await self.session.clear()
        self.send_response({'loggedOut': True})


class ConnectionRequestHandler(base.AuthenticatedRequestHandler):
    async def delete(self, integration_name: str) -> None:
        await self.postgres_execute(
            'DELETE FROM v1.user_oauth2_tokens'
            ' WHERE integration = %(integration_name)s'
            '   AND username = %(username)s', {
                'integration_name': integration_name,
                'username': self._current_user.username,
            })
        self.set_status(204)



from __future__ import annotations

import re

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):
    NAME = 'notifications'
    ITEM_NAME = 'notification'
    ID_KEY = ['integration_name', 'name']
    FIELDS = []

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT notification_name AS name, integration_name, id_pattern,
               documentation, default_action, created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.integration_notifications
         WHERE integration_name = %(integration_name)s
         ORDER BY name
        """)

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT notification_name AS name, integration_name, id_pattern,
               documentation, default_action, created_at, created_by,
               last_modified_at, last_modified_by
          FROM v1.integration_notifications
         WHERE integration_name = %(integration_name)s
           AND notification_name = %(name)s
        """)

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.integration_notifications
                    (notification_name, integration_name, id_pattern,
                     documentation, default_action, verification_token,
                     created_at, created_by)
             VALUES (%(name)s, %(integration_name)s, %(id_pattern)s,
                     %(documentation)s, %(default_action)s,
                     %(verification_token)s, CURRENT_TIMESTAMP, %(username)s)
          RETURNING notification_name AS name, integration_name, id_pattern,
                    documentation, default_action
        """)

    @base.require_permission('admin')
    async def post(self, *args, **kwargs) -> None:
        await super().post(*args, **kwargs)


class RecordRequestHandler(base.CRUDRequestHandler):
    NAME = 'notification'
    ID_KEY = ['integration_name', 'name']
    OMIT_FIELDS = ['verification_token']

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT notification_name AS name, integration_name, id_pattern,
               documentation, verification_token, default_action,
               created_at, created_by, last_modified_at, last_modified_by
          FROM v1.integration_notifications
         WHERE notification_name = %(name)s
           AND integration_name = %(integration_name)s
        """)

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.integration_notifications
           SET default_action = %(default_action)s,
               documentation = %(documentation)s,
               id_pattern = %(id_pattern)s,
               last_modified_at = CURRENT_TIMESTAMP,
               last_modified_by = %(username)s,
               notification_name = %(name)s,
               verification_token = %(verification_token)s
         WHERE notification_name = %(current_name)s
           AND integration_name = %(integration_name)s
        """)

    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.integration_notifications
         WHERE notification_name = %(name)s
           AND integration_name = %(integration_name)s
        """)

    @base.require_permission('admin')
    async def delete(self, *args, **kwargs) -> None:
        await super().delete(*args, **kwargs)

    @base.require_permission('admin')
    async def patch(self, *args, **kwargs) -> None:
        await super().patch(*args, **kwargs)



import ldap3
from tornado import testing

from imbi import ldap
from tests import base

LDAP_USER = 'test'
LDAP_PASSWORD = 'password'


class ClientTestCase(base.TestCase):
    @staticmethod
    def get_user_dn(conn):
        result = conn.extend.standard.who_am_i()
        return result[3:]

    @testing.gen_test
    async def test_attributes(self):
        client = ldap.Client(self.settings['ldap'])
        conn = await client.connect(LDAP_USER, LDAP_PASSWORD)
        self.assertIsInstance(conn, ldap3.Connection)
        attributes = await client.attributes(conn, self.get_user_dn(conn))
        self.assertEqual(attributes['givenName'], 'Imbi')
        self.assertEqual(attributes['sn'], 'Test')
        self.assertEqual(attributes['mail'], 'imbi@example.org')
        conn.unbind()

    @testing.gen_test
    async def test_groups(self):
        client = ldap.Client(self.settings['ldap'])
        conn = await client.connect(LDAP_USER, LDAP_PASSWORD)
        self.assertIsInstance(conn, ldap3.Connection)
        groups = await client.groups(conn, self.get_user_dn(conn))
        self.assertIn('cn=imbi,ou=groups,dc=example,dc=org', groups)
        conn.unbind()



import collections.abc
import decimal
import operator
import typing

import pydantic

from imbi import errors, models
from imbi.endpoints import base

# makes pydantic serialize decimal.Decimal as a float instead of a string
JsonDecimal = typing.Annotated[decimal.Decimal,
                               pydantic.PlainSerializer(
                                   lambda x: float(x),
                                   return_type=float,
                                   when_used='json',
                               )]


class ScoreDetail(pydantic.BaseModel):
    name: str
    fact_type: str
    data_type: str
    value: typing.Union[str, bool, float, int, None]
    score: JsonDecimal
    weight: int


class EnumValue(pydantic.BaseModel):
    value: str
    score: JsonDecimal
    selected: bool


class EnumScoreDetail(ScoreDetail):
    enums: list[EnumValue]


class RangeValue(pydantic.BaseModel):
    value: tuple[float, float]
    score: JsonDecimal
    selected: bool


class RangeScoreDetail(ScoreDetail):
    ranges: list[RangeValue]


FactId = int
FactRow = collections.abc.Mapping[str, typing.Union[int, str, float]]
FactMap = collections.abc.Mapping[FactId, collections.abc.Sequence[FactRow]]


class ProjectScoreDetailHandler(base.AuthenticatedRequestHandler):
    """Implementation of the v1.project_score() stored proc in Python

    The stored procedure is the "source of truth" for the project
    score. This is a facsimile to explain what goes in to the score
    calculation and how changes could affect the score.

    """
    NAME = 'project-score-detail'

    async def get(self, project_id: int) -> None:
        result = await self.postgres_execute(
            'SELECT project_type_id'
            '  FROM v1.projects'
            ' WHERE id = %(project_id)s', {'project_id': project_id})
        if not result:
            raise errors.ItemNotFound()

        facts = await models.project_facts(project_id, self.application)
        enum_details = await self._gather_facts(
            [f.id for f in facts if f.fact_type == 'enum'],
            'SELECT fact_type_id, value, score'
            '  FROM v1.project_fact_type_enums'
            ' WHERE fact_type_id IN %(fact_type_ids)s',
            'score',
            'value',
        )
        range_details = await self._gather_facts(
            [f.id for f in facts if f.fact_type == 'range'],
            'SELECT fact_type_id, score, min_value, max_value'
            '  FROM v1.project_fact_type_ranges'
            ' WHERE fact_type_id IN %(fact_type_ids)s',
            'score',
        )
        scored_facts = sorted((f for f in facts if f.weight),
                              key=operator.attrgetter('weight'),
                              reverse=True)

        # imbi.models sets the process-wide precision to 2 but
        # the database calculation uses decimal(9, 2) so we need
        # to use a matching precision
        with decimal.localcontext() as ctx:
            ctx.prec = 9
            scored_values = self._generate_score_detail(
                scored_facts, enum_details, range_details)
            total_score = sum(value.weight * value.score
                              for value in scored_values)
            total_weight = sum(f.weight for f in scored_facts)
            self.send_response({
                'score': total_score / total_weight,
                'scored_facts': scored_values,
                'unscored_facts': [{
                    'name': f.name,
                    'fact_type': f.fact_type,
                    'data_type': f.data_type,
                    'score': None,
                    'value': f.value,
                    'weight': 0,
                } for f in facts if not f.weight]
            })

    async def _gather_facts(self, fact_ids: collections.abc.Sequence[int],
                            query: str, *sort_params: str) -> FactMap:
        details = collections.defaultdict[int, list[dict]](list)
        if fact_ids:
            result = await self.postgres_execute(
                query, {'fact_type_ids': tuple(fact_ids)})
            for row in result:
                details[row['fact_type_id']].append(row)
        for value in details.values():
            value.sort(key=operator.itemgetter(*sort_params), reverse=True)
        return details

    @staticmethod
    def _generate_score_detail(
        scored_facts: typing.Sequence[models.ProjectFact],
        enum_details,
        range_details,
    ) -> list['ScoreDetail']:
        scored_values = []
        for fact in scored_facts:
            if fact.data_type == 'boolean':
                score = decimal.Decimal(100.0 if fact.value else 0)
            else:
                score = fact.score

            score_detail: ScoreDetail = ScoreDetail(
                name=fact.name,
                fact_type=fact.fact_type,
                data_type=fact.data_type,
                value=fact.value,
                score=score,
                weight=fact.weight,
            )
            if fact.fact_type == 'range':
                ranges = [
                    RangeValue(
                        value=(r['min_value'], r['max_value']),
                        score=r['score'],
                        selected=(r['min_value'] <
                                  (fact.value or 0) <= r['max_value']),
                    ) for r in range_details[fact.id]
                ]
                score_detail = RangeScoreDetail.model_validate(
                    score_detail.model_dump() | {'ranges': ranges})
            elif fact.fact_type == 'enum':
                enums = [
                    EnumValue(value=e['value'],
                              score=e['score'],
                              selected=fact.value == e['value'])
                    for e in enum_details[fact.id]
                ]
                score_detail = EnumScoreDetail.model_validate(
                    score_detail.model_dump() | {'enums': enums})
            scored_values.append(score_detail)

        return scored_values



"""
Some postgres/sql/psycopg helper functions

I would have called this module either psycopg or sql but both
names can be pretty problematic, so I settled on postgres. The
goal is to collect SQL query helpers into this module when they
are applicable to multiple uses.

"""
from __future__ import annotations

import typing
import uuid

import sprockets_postgres
from psycopg2 import sql

_omitted = str(uuid.uuid4())


async def insert_values(
        conn: sprockets_postgres.PostgresConnector, schema: str,
        table_name: str, column_names: typing.Iterable[str],
        rows: typing.Iterable[typing.Iterable[typing.Any]]) -> None:
    """Insert multiple rows at the same time

    This is similar to the psycopg2 execute_values function and
    I would use it were it included in psycopg 3. Instead, this
    implements the SQL building approach described briefly in
    https://github.com/psycopg/psycopg/issues/576 .

    .. code-block::

       INSERT INTO {schema}.{table_name} ({column_names})
            VALUES (*rows[0]), (*rows[1]), ...

    """
    if not rows:
        return

    # please avert your eyes ... I apologize
    query = sql.SQL('INSERT INTO {table} ({columns}) VALUES {values}').format(
        table=sql.Identifier(schema, table_name),
        columns=sql.SQL(',').join(sql.Identifier(n) for n in column_names),
        values=sql.SQL(',').join(
            sql.SQL('({})').format(
                sql.SQL(',').join(sql.Literal(v) for v in row))
            for row in rows))
    await conn.execute(query.as_string(conn.cursor.raw),
                       metric_name=f'insert-{table_name}')


async def update_entity(
    conn: sprockets_postgres.PostgresConnector,
    schema: str,
    table_name: str,
    original: dict[str, typing.Any],
    updated: dict[str, typing.Any],
    columns: typing.Iterable[str],
    *,
    id_column: str = 'id',
    id_value: typing.Any = _omitted,
) -> None:
    """Update an entity from `original` into `updated`

    This function dynamically generates the SQL UPDATE statement
    constraining the changes to those in `columns`. The row is
    assumed to be identified by a column named ``id``. The value
    is taken from ``original['id']``.

    .. code-block::

       UPDATE {schema}.{table_name}
          SET col1 = updated[col1],
              ...
        WHERE id = original[id]

    """
    modified_columns = [c for c in columns if original[c] != updated[c]]
    if not modified_columns:
        return None

    id_value = original[id_column] if id_value is _omitted else id_value
    query = sql.SQL(
        'UPDATE {table} SET {values} WHERE {id_column} = {id_value}').format(
            table=sql.Identifier(schema, table_name),
            values=sql.SQL(',').join(
                sql.SQL('{} = {}').format(sql.Identifier(column),
                                          sql.Literal(updated[column]))
                for column in modified_columns),
            id_column=sql.Identifier(id_column),
            id_value=sql.Literal(id_value),
        )
    await conn.execute(query.as_string(conn.cursor.raw),
                       metric_name=f'update-{table_name}')



from __future__ import annotations

import datetime
import enum
import functools
import typing

import pydantic

BOMRef = typing.Annotated[str, pydantic.StringConstraints(min_length=1)]
PackageURL = typing.Annotated[str,
                              pydantic.StringConstraints(pattern=r'^pkg:')]
VersionMap = typing.MutableMapping[str, int]


class Dependency(pydantic.BaseModel):
    ref: BOMRef
    depends_on: list[BOMRef] = pydantic.Field(default_factory=list,
                                              alias='dependsOn')


class ExternalReference(pydantic.BaseModel):
    url: str
    type: str


class Component(pydantic.BaseModel):
    name: str
    bom_ref: typing.Union[BOMRef, None] = pydantic.Field(default=None,
                                                         alias='bom-ref')
    purl: typing.Union[PackageURL, None] = pydantic.Field(default=None,
                                                          frozen=True)
    version: typing.Union[str, None] = None
    external_references: list[ExternalReference] = pydantic.Field(
        default_factory=list, alias='externalReferences')

    @functools.cached_property
    def package_purl(self) -> typing.Union[str, None]:
        """Package URL without trailing info (eg, version, fragment, etc)"""
        return self.purl.partition('@')[0] if self.purl else None

    @property
    def home_page(self) -> typing.Union[str, None]:
        refs: dict[str, str] = {}
        for reference in self.external_references:
            refs[reference.type] = reference.url
        for ref_type in ['website', 'documentation', 'release-notes']:
            if page := refs.get(ref_type):
                return page
        return None


class SBOMMetadata(pydantic.BaseModel):
    component: typing.Union[Component, None] = None
    timestamp: typing.Union[datetime.datetime, None] = None


class SBOMSpecVersion(str, enum.Enum):
    V_1_5 = '1.5'
    V_1_6 = '1.6'


class SBOM(pydantic.BaseModel):
    bom_format: str = pydantic.Field(alias='bomFormat', pattern='CycloneDX')
    spec_version: SBOMSpecVersion = pydantic.Field(alias='specVersion')
    metadata: SBOMMetadata = pydantic.Field(default_factory=SBOMMetadata)
    components: list[Component] = pydantic.Field(default_factory=list)
    dependencies: list[Dependency] = pydantic.Field(default_factory=list)



from __future__ import annotations

import logging
import typing

import pydantic
import sprockets.mixins.http
import yarl

from imbi import errors, models, version
if typing.TYPE_CHECKING:  # pragma: nocover
    from imbi import app


class ServiceInfo(pydantic.BaseModel):
    html_url: pydantic.HttpUrl  # aweber.pagerduty.com/...
    id: str
    self: pydantic.HttpUrl  # api.pagerduty.com/...


class _CreateServiceResponse(pydantic.BaseModel):
    service: ServiceInfo


class IntegrationInfo(pydantic.BaseModel):
    id: str
    name: str
    type: str
    self: pydantic.HttpUrl
    html_url: pydantic.HttpUrl
    integration_key: str


class _CreateIntegrationResponse(pydantic.BaseModel):
    integration: IntegrationInfo


class RelationshipServiceInfo(pydantic.BaseModel):
    id: str
    type: str


class ServiceRelationship(pydantic.BaseModel):
    supporting_service: RelationshipServiceInfo
    dependent_service: RelationshipServiceInfo
    id: str
    type: str


class _GetServiceDependenciesResponse(pydantic.BaseModel):
    relationships: list[ServiceRelationship]


class _PagerDutyClient(sprockets.mixins.http.HTTPClientMixin):
    def __init__(self, info: models.Integration, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__package__).getChild(
            'PagerDutyClient')
        self.api_url = yarl.URL(str(info.api_endpoint))
        self.api_secret = info.api_secret
        self.headers = {
            'Accept': 'application/json',
            'Authorization': f'Token token={info.api_secret}',
        }

    async def api(self,
                  url: yarl.URL | str,
                  *,
                  method: str = 'GET',
                  **kwargs) -> sprockets.mixins.http.HTTPResponse:
        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        url = self.api_url / url.path.removeprefix('/')

        request_headers = kwargs.pop('request_headers', {})
        request_headers.update(self.headers)
        kwargs['request_headers'] = request_headers

        if kwargs.get('body') is not None:
            request_headers['Content-Type'] = 'application/json'
            kwargs['content_type'] = 'application/json'
        kwargs['user_agent'] = f'imbi/{version} (PagerDutyClient)'
        rsp = await self.http_fetch(str(url), method=method, **kwargs)
        if not rsp.ok:
            self.logger.error('%s %s failed: %s', method, url, rsp.code)
            self.logger.debug('pagerduty key: %s...%s', self.api_secret[:4],
                              self.api_secret[-4:])
            self.logger.debug('response body: %r', rsp.body)
        return rsp

    async def create_service(self, project: models.Project) -> ServiceInfo:
        """Create a PagerDuty service for `project`

        The new service is named `project.slug` and assigned to the
        escalation policy from `project.namespace.pagerduty_policy`.
        If the namespaces pagerduty policy is NULL, then an
        ``InternalServiceError`` is raised.
        """
        if project.namespace.pagerduty_policy is None:
            raise errors.InternalServerError(
                'PagerDuty is not enabled for namespace %s',
                project.namespace.slug)
        body = {
            'name': project.slug,
            'escalation_policy': {
                'id': project.namespace.pagerduty_policy,
                'type': 'escalation_policy_reference'
            }
        }
        if project.description:
            body['description'] = project.description
        self.logger.debug('creating PagerDuty service with: %r', body)
        response = await self.api('/services', method='POST', body=body)
        if not response.ok:
            self.logger.error('PagerDuty API failure: body %r', response.body)
            raise errors.InternalServerError(
                'POST %s failed for name=%r policy=%r: %s',
                response.history[-1].effective_url,
                project.slug,
                project.namespace.pagerduty_policy,
                response.code,
                title='PagerDuty API Failure',
                detail=f'Failed to create service {body["name"]!r}')

        try:
            parsed_rsp = _CreateServiceResponse.model_validate(response.body)
        except pydantic.ValidationError as error:
            raise errors.InternalServerError(
                'Failed to parse the PagerDuty response: %s',
                error,
                title='PagerDuty API Failure',
                detail='Failed to parse service creation response') from None
        else:
            return parsed_rsp.service

    async def create_inbound_api_integration(
            self, name: str, service: ServiceInfo) -> IntegrationInfo:
        response = await self.api(
            yarl.URL('/services') / service.id / 'integrations',
            method='POST',
            body={
                'integration': {
                    'name': name,
                    'type': 'generic_events_api_inbound_integration'
                }
            },
        )
        if not response.ok:
            self.logger.error('PagerDuty API failure: body %r', response.body)
            raise errors.InternalServerError(
                'POST %s failed: %s',
                response.history[-1].effective_url,
                response.code,
                title='PagerDuty API Failure',
                detail=f'Failed to create inbound API integration {name}')

        try:
            parsed_rsp = _CreateIntegrationResponse.model_validate(
                response.body)
        except pydantic.ValidationError as error:
            raise errors.InternalServerError(
                'Failed to parse the PagerDuty response: %s',
                error,
                title='PagerDuty API Failure',
                detail='Failed to parse integration creation response'
            ) from None
        else:
            return parsed_rsp.integration

    async def remove_service(self, service_id: str) -> None:
        """Delete `service_id` from PagerDuty"""
        await self.api(yarl.URL('/services') / service_id, method='DELETE')

    async def get_service_dependencies(
            self, service_id) -> list[ServiceRelationship]:
        self.logger.debug('fetching PagerDuty service dependencies for %s',
                          service_id)
        response = await self.api(
            f'/service_dependencies/technical_services/{service_id}')
        if not response.ok:
            raise errors.InternalServerError(
                'GET %s failed to fetch service dependencies for %s: %s',
                response.history[-1].effective_url,
                service_id,
                response.code,
                title='PagerDuty API Failure',
                detail=f'Failed to fetch service dependencies for {service_id}'
            )
        try:
            parsed_rsp = _GetServiceDependenciesResponse.model_validate(
                response.body)
        except pydantic.ValidationError as error:
            raise errors.InternalServerError(
                'Failed to parse the PagerDuty response: %s',
                error,
                title='PagerDuty API Failure',
                detail='Failed to parse fetch service dependencies response'
            ) from None
        else:
            return parsed_rsp.relationships

    async def associate_service_dependency(self, supporting_service_id: str,
                                           dependent_service_id: str) -> None:
        """Create a dependency between two services"""
        body = {
            'relationships': [{
                'supporting_service': {
                    'id': supporting_service_id,
                    'type': 'service',
                },
                'dependent_service': {
                    'id': dependent_service_id,
                    'type': 'service',
                }
            }]
        }
        self.logger.debug('creating PagerDuty service dependency with: %r',
                          body)
        response = await self.api('/service_dependencies/associate',
                                  method='POST',
                                  body=body)
        if not response.ok:
            raise errors.InternalServerError(
                'POST %s failed for service association %s to %s: %s',
                response.history[-1].effective_url,
                supporting_service_id,
                dependent_service_id,
                response.code,
                title='PagerDuty API Failure',
                detail=f'Failed to associate service {supporting_service_id} '
                f'as dependency of service {dependent_service_id}')

    async def disassociate_service_dependency(
            self, supporting_service_id: str,
            dependent_service_id: str) -> None:
        """Remove a dependency between two services"""
        body = {
            'relationships': [{
                'supporting_service': {
                    'id': supporting_service_id,
                    'type': 'service',
                },
                'dependent_service': {
                    'id': dependent_service_id,
                    'type': 'service',
                }
            }]
        }
        self.logger.debug('removing PagerDuty service dependency with: %r',
                          body)
        response = await self.api('/service_dependencies/disassociate',
                                  method='POST',
                                  body=body)
        if not response.ok:
            raise errors.InternalServerError(
                'POST %s failed for service disassociation %s to %s: %s',
                response.history[-1].effective_url,
                supporting_service_id,
                dependent_service_id,
                response.code,
                title='PagerDuty API Failure',
                detail='Failed to disassociate service '
                f'{supporting_service_id} as dependency of service '
                f'{dependent_service_id}')


async def create_client(application: app.Application,
                        integration_name: str) -> _PagerDutyClient:
    """Create a PagerDuty API client

    The settings need to create a client are stored in the database.
    However, you can disable the PagerDuty client by setting the
    `/automations/pagerduty/enabled` property to false in the
    configuration file.

    If the automation is disabled in the configuration file, the
    integration is missing from the database, or the integration
    does not have an API secret set, then a ``ClientUnavailableError``
    is raised with an appropriate diagnostic message.
    """
    logger = logging.getLogger(__package__).getChild('create_client')
    try:
        settings = application.settings['automations']['pagerduty']
        enabled = settings['enabled']
    except KeyError:
        raise errors.ClientUnavailableError(integration_name, 'not configured')
    else:
        if not enabled:
            raise errors.ClientUnavailableError(integration_name, 'disabled')

    pagerduty_info = await models.integration(integration_name, application)
    if not pagerduty_info:
        logger.warning('%r integration is enabled by not configured',
                       integration_name)
        raise errors.ClientUnavailableError(integration_name, 'not configured')
    if not pagerduty_info.api_secret:
        logger.warning('API secret is missing for %r', integration_name)
        raise errors.ClientUnavailableError(integration_name, 'misconfigured')

    return _PagerDutyClient(pagerduty_info)



import re

from imbi.endpoints import base, projects


class _RequestMixin:

    ID_KEY = ['project_id', 'environment']
    ITEM_NAME = 'project-url'
    FIELDS = ['project_id', 'environment', 'url']
    TTL = 300

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT project_id,
               environment,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               url
          FROM v1.project_urls
         WHERE project_id = %(project_id)s
           AND environment = %(environment)s""")


class CollectionRequestHandler(projects.ProjectAttributeCollectionMixin,
                               _RequestMixin, base.CollectionRequestHandler):

    NAME = 'project-urls'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT project_id,
               environment,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               url
          FROM v1.project_urls
         WHERE project_id = %(project_id)s
         ORDER BY environment""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.project_urls
                    (project_id, environment, created_by, url)
             VALUES (%(project_id)s, %(environment)s, %(username)s, %(url)s)
          RETURNING project_id, environment""")


class RecordRequestHandler(projects.ProjectAttributeCRUDMixin, _RequestMixin,
                           base.CRUDRequestHandler):

    NAME = 'project-url'

    DELETE_SQL = re.sub(
        r'\s+', ' ', """\
        DELETE FROM v1.project_urls
              WHERE project_id = %(project_id)s
                AND environment = %(environment)s""")

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.project_urls
           SET url=%(url)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s
         WHERE project_id = %(project_id)s
           AND environment = %(environment)s""")



"""
Application Metrics

"""
import collections
import logging
import re
import socket

from . import base

LOGGER = logging.getLogger(__name__)
INF = float('inf')
TAGS = re.compile(r'(([\w_-]+)=([\w_\+\.-]+))')


class RequestHandler(base.RequestHandler):
    """Returns internal metrics in Prometheus line format"""
    BUCKETS = (.005, .01, .025, .05, .075, .1, .25, .5, .75, 1.0, 2.5, 5.0,
               7.5, 10.0, INF)
    NAME = 'metrics'

    @staticmethod
    def _build_key_dict(key_in: str) -> dict:
        tags = {}
        for match in TAGS.finditer(key_in):
            tags[match.group(2)] = match.group(3)
        return tags

    def _build_output_key(self, key_in: str, suffix: str) -> str:
        tags = self._build_key_dict(key_in)
        key = tags.pop('key')
        tag_repr = ','.join(f'{k}="{v}"' for k, v in tags.items())
        return f'imbi_{key}_{suffix}{{{tag_repr}}}'

    async def get(self):
        """Tornado RequestHandler GET request endpoint for reporting status"""
        all_hosts = self.get_argument('all_hosts', 'false').lower() == 'true'
        flush = self.get_argument('flush', 'true').lower() == 'true'

        self.set_header('Content-Type', 'text/plain; version=0.0.4')
        self.write('# Prometheus line format\n')
        self.write(
            f'# Generated on {socket.gethostname()}, all_hosts={all_hosts},'
            f' flush={flush}\n\n')

        counters = await self.application.stats.counters(all_hosts, flush)
        chunks = {}
        for counter in sorted(counters.keys()):
            key = self._build_output_key(counter, 'total')
            chunk_key = key.split('{')[0]
            if chunk_key not in chunks:
                chunks[chunk_key] = []
            chunks[chunk_key].append(f'{key} {counters[counter]}\n')
        for key in chunks:
            self.write(f'# TYPE {key} counter\n')
            for item in chunks[key]:
                self.write(item)
            self.write('\n')

        timers = await self.application.stats.durations(all_hosts, flush)
        values: dict[str, list[str]] = {}
        for timer in sorted(timers.keys()):
            chunk_key = self._build_output_key(timer, 'seconds').split('{')[0]
            if chunk_key not in values:
                values[chunk_key] = []
            buckets = collections.Counter()
            counts = collections.Counter()
            sums = collections.Counter()
            while timers[timer]:
                value = timers[timer].pop(0)
                counts[timer] += 1
                sums[timer] += value
                for bucket in self.BUCKETS:
                    if bucket not in buckets:
                        buckets[bucket] = 0
                    if value <= bucket:
                        buckets[bucket] += 1
            for bucket in self.BUCKETS:
                if bucket == INF:
                    continue
                key = self._build_output_key(f'{timer}:le={bucket}',
                                             'seconds_bucket')
                values[chunk_key].append(f'{key} {buckets[bucket]}\n')

            key = self._build_output_key(f'{timer}:le=Inf+', 'seconds_bucket')
            values[chunk_key].append(f'{key} {counts[timer]}\n')

            key = self._build_output_key(f'{timer}', 'seconds_count')
            values[chunk_key].append(f'{key} {counts[timer]}\n')
            key = self._build_output_key(f'{timer}', 'seconds_sum')
            values[chunk_key].append(f'{key} {sums[timer]}\n')

        for chunk_key in sorted(values):
            self.write(f'# TYPE {chunk_key} histogram\n')
            for metric in values[chunk_key]:
                self.write(metric)
            self.write('\n')

        postgres = await self.application.postgres_status()
        self.write('# HELP postgres_pool_size The number of open connections '
                   'in the pool\n')
        self.write('# TYPE postgres_pool_size gauge\n')
        self.write(f'imbi_postgres_pool_size{{host="{socket.gethostname()}"}} '
                   f'{postgres["pool_size"]}\n\n')
        self.write('# HELP postgres_pool_free The number of free connections '
                   'in the pool\n')
        self.write('# TYPE postgres_pool_free gauge\n')
        self.write(f'imbi_postgres_pool_free{{host="{socket.gethostname()}"}} '
                   f'{postgres["pool_free"]}\n\n')



import aioboto3


async def get_credentials(session: aioboto3.Session, role_arn: str,
                          username: str, id_token: str):
    """Fetch temporary AWS credentials from STS."""
    async with session.client('sts') as client:
        return await client.assume_role_with_web_identity(
            RoleArn=role_arn,
            RoleSessionName=f'{username}-imbi',
            WebIdentityToken=id_token)


async def get_parameter(name: str, session: aioboto3.Session,
                        access_key_id: str, secret_access_key: str,
                        session_token: str) -> dict:
    """Fetch the SSM parameter at the specified name."""
    async with session.client('ssm',
                              aws_access_key_id=access_key_id,
                              aws_secret_access_key=secret_access_key,
                              aws_session_token=session_token) as client:
        parameter = await client.get_parameter(Name=name, WithDecryption=True)
        return parameter['Parameter']


async def get_parameters_by_path(session: aioboto3.Session, path: str,
                                 access_key_id: str, secret_access_key: str,
                                 session_token: str) -> list:
    """Fetch all SSM parameters under the specified path."""
    async with session.client('ssm',
                              aws_access_key_id=access_key_id,
                              aws_secret_access_key=secret_access_key,
                              aws_session_token=session_token) as client:

        params = []
        paginator = client.get_paginator('get_parameters_by_path')
        async for page in paginator.paginate(Path=path,
                                             Recursive=True,
                                             WithDecryption=True,
                                             MaxResults=10):
            params.extend(page['Parameters'])
        return params


async def put_parameter(session: aioboto3.Session,
                        access_key_id: str,
                        secret_access_key: str,
                        session_token: str,
                        path: str,
                        value: str,
                        param_type: str = None,
                        overwrite: bool = False):
    kwargs = {
        'Name': path,
        'Value': value,
        'Overwrite': overwrite,
    }
    if param_type is not None:
        kwargs['Type'] = param_type

    async with session.client('ssm',
                              aws_access_key_id=access_key_id,
                              aws_secret_access_key=secret_access_key,
                              aws_session_token=session_token) as client:
        await client.put_parameter(**kwargs)


async def delete_parameter(session: aioboto3.Session, access_key_id: str,
                           secret_access_key: str, session_token: str,
                           path: str):
    async with session.client('ssm',
                              aws_access_key_id=access_key_id,
                              aws_secret_access_key=secret_access_key,
                              aws_session_token=session_token) as client:
        await client.delete_parameter(Name=path)



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = 'id'
    FIELDS = [
        'id', 'name', 'slug', 'icon_class', 'maintained_by',
        'gitlab_group_name', 'sentry_team_slug', 'pagerduty_policy',
        'aws_ssm_slug'
    ]
    DEFAULTS = {'icon_class': 'fas fa-users', 'maintained_by': []}

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT id, "name", created_at, created_by,
               last_modified_at, last_modified_by,
               slug, icon_class, maintained_by,
               gitlab_group_name, sentry_team_slug,
               pagerduty_policy, aws_ssm_slug
          FROM v1.namespaces WHERE id=%(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'namespaces'
    ITEM_NAME = 'namespace'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """ \
        SELECT id, "name", slug, icon_class, maintained_by, gitlab_group_name,
               sentry_team_slug, pagerduty_policy, aws_ssm_slug
          FROM v1.namespaces ORDER BY "name" ASC""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.namespaces
                    ("name", created_by, slug, icon_class, "maintained_by",
                     gitlab_group_name, sentry_team_slug, pagerduty_policy,
                     aws_ssm_slug)
             VALUES (%(name)s, %(username)s, %(slug)s, %(icon_class)s,
                     %(maintained_by)s, %(gitlab_group_name)s,
                     %(sentry_team_slug)s, %(pagerduty_policy)s,
                     %(aws_ssm_slug)s)
          RETURNING id""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'namespace'

    DELETE_SQL = 'DELETE FROM v1.namespaces WHERE id=%(id)s'

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.namespaces
           SET "name" = %(name)s,
               last_modified_at = CURRENT_TIMESTAMP,
               last_modified_by = %(username)s,
               slug = %(slug)s,
               icon_class = %(icon_class)s,
               "maintained_by" = %(maintained_by)s,
               gitlab_group_name = %(gitlab_group_name)s,
               sentry_team_slug = %(sentry_team_slug)s,
               pagerduty_policy = %(pagerduty_policy)s,
               aws_ssm_slug = %(aws_ssm_slug)s
         WHERE id=%(id)s""")
